# -*- coding: utf-8 -*-
__author__ = "Benedict J. Heinen"
__copyright__ = "Copyright 2021, Benedict J. Heinen"
__email__ = "benedict.heinen@gmail.com"

import os.path
import datetime
import numpy as np
import lmfit
from PyQt5.QtCore import Qt, pyqtSignal, QSignalMapper, QObject
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import QWidget, QFrame, QGridLayout, QVBoxLayout, \
                            QHBoxLayout, QGroupBox, QPushButton, QRadioButton, \
                            QLineEdit, QDoubleSpinBox, QLabel, QScrollArea, \
                            QCheckBox, QSplitter, QComboBox, QStyle
# LiquidDiffract imports
from LiquidDiffract.gui import plot_widgets
from LiquidDiffract.core import data_utils
from LiquidDiffract.core import peak_fit
import LiquidDiffract.core.core as core
from LiquidDiffract.version import __appname__, __version__


class StructureUI(QWidget):

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)

        # Make config widget
        self.structure_config_widget = StructureConfigWidget()
        self.config_scroll_area = QScrollArea()
        self.config_scroll_area.setFrameShape(QFrame.NoFrame)
        self.config_scroll_area.setWidget(self.structure_config_widget)
        self.config_scroll_area.setWidgetResizable(True)

        # Make vertical line separator
        self.vline = QFrame()
        self.vline.setFrameShape(QFrame.VLine)
        self.vline.setFrameShadow(QFrame.Sunken)
        self.vline.setObjectName('vline')

        # Make plot widget
        self.structure_plot_widget = plot_widgets.StructurePlotWidget()
        self.plot_scroll_area = QScrollArea()
        self.plot_scroll_area.setWidget(self.structure_plot_widget)
        self.plot_scroll_area.setWidgetResizable(True)
        self.plot_scroll_area.setFrameShape(QFrame.NoFrame)

        # Create resizeable splitter
        self.hsplitter = QSplitter(Qt.Horizontal)
        self.hsplitter.addWidget(self.config_scroll_area)
        self.hsplitter.addWidget(self.plot_scroll_area)
        self.hsplitter.setStretchFactor(0, 2)
        self.hsplitter.setStretchFactor(1, 5)

        self.layout.addWidget(self.hsplitter)

        self.setLayout(self.layout)

        # Initialise data
        self.active_func = 'rdf'
        self.clear_data()

        self.create_signals()

    def create_signals(self):

        self.structure_config_widget.plot_view_gb.rdf_btn.toggled.connect(self.toggle_plot_view)

        self.structure_config_widget.monatomic_gb.r0_input.valueChanged.connect(self.update_plot_data)
        self.structure_config_widget.monatomic_gb.rpmax_input.valueChanged.connect(self.update_plot_data)
        self.structure_config_widget.monatomic_gb.rmax_input.valueChanged.connect(self.update_plot_data)
        self.structure_config_widget.monatomic_gb.rmin_input.valueChanged.connect(self.update_plot_data)

        self.structure_config_widget.monatomic_gb.r0_input.valueChanged.connect(self.calc_integrals)
        self.structure_config_widget.monatomic_gb.rpmax_input.valueChanged.connect(self.calc_integrals)
        self.structure_config_widget.monatomic_gb.rmax_input.valueChanged.connect(self.calc_integrals)
        self.structure_config_widget.monatomic_gb.rmin_input.valueChanged.connect(self.calc_integrals)

        self.int_limit_signalMapper = QSignalMapper()
        self.int_limit_signalMapper.mapped[QObject].connect(self.update_int_limits)

        self.structure_config_widget.monatomic_gb.find_limits_btn.clicked.connect(self.auto_integration_limits)
        self.structure_config_widget.monatomic_gb.calc_N_btn.clicked.connect(self.calc_integrals)

        self.structure_config_widget.monatomic_gb.toggled.connect(self.toggle_monatomic_gb)
        self.structure_config_widget.polyatomic_gb.toggled.connect(self.toggle_polyatomic_gb)

        self.structure_plot_widget.fit_limits.sigRegionChanged.connect(self.update_fit_limits_sb)
        self.structure_config_widget.polyatomic_gb.min_limit_input.valueChanged.connect(self.update_fit_limits_plot)
        self.structure_config_widget.polyatomic_gb.max_limit_input.valueChanged.connect(self.update_fit_limits_plot)

        self.structure_config_widget.polyatomic_gb.fit_peaks_btn.clicked.connect(self.on_click_fit_peaks)
        self.structure_config_widget.polyatomic_gb.guess_peak.connect(self.guess_peak_params)
        self.structure_config_widget.polyatomic_gb.gauss_params_changed_relay.connect(self.make_peak_plots)

    def create_int_limit_signals(self):
        try:
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.r0_line_rdf, self.structure_plot_widget.r0_line_rdf)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.r0_line_tr, self.structure_plot_widget.r0_line_tr)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.rpmax_line_rdf, self.structure_plot_widget.rpmax_line_rdf)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.rpmax_line_tr, self.structure_plot_widget.rpmax_line_tr)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.rmax_line_rdf, self.structure_plot_widget.rmax_line_rdf)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.rmax_line_tr, self.structure_plot_widget.rmax_line_tr)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.rmin_line_rdf, self.structure_plot_widget.rmin_line_rdf)
            self.int_limit_signalMapper.setMapping(self.structure_plot_widget.rmin_line_tr, self.structure_plot_widget.rmin_line_tr)

            self.structure_plot_widget.r0_line_rdf.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.r0_line_tr.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.rpmax_line_rdf.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.rpmax_line_tr.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.rmax_line_rdf.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.rmax_line_tr.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.rmin_line_rdf.sigDragged.connect(self.int_limit_signalMapper.map)
            self.structure_plot_widget.rmin_line_tr.sigDragged.connect(self.int_limit_signalMapper.map)
        except AttributeError:
            return

    def update_int_limits(self, _sender):

        _new_pos = _sender.value()
        _limit = _sender.name()
        if _limit == 'r0':
            self.structure_config_widget.monatomic_gb.r0_input.setValue(_new_pos)
        elif _limit == 'rpmax':
            self.structure_config_widget.monatomic_gb.rpmax_input.setValue(_new_pos)
        elif _limit == 'rmax':
            self.structure_config_widget.monatomic_gb.rmax_input.setValue(_new_pos)
        elif _limit == 'rmin':
            self.structure_config_widget.monatomic_gb.rmin_input.setValue(_new_pos)

        else:
            raise ValueError

    def clear_data(self):

        self.data = {'rdf_x': np.asarray([]), 'rdf_y': np.asarray([]),
                     'tr_x': np.asarray([]), 'tr_y': np.asarray([]),
                     'r0': 0.0, 'rmax': 0.0, 'rpmax': 0.0, 'rmin': 0.0,
                     'fit_r': np.asarray([]),
                     'obj_fun': np.asarray([]),
                     'gauss_peaks': [],
                     'gauss_model': np.asarray([]),
                     'gauss_residuals': np.asarray([]),
                     'gauss_residuals_full': np.asarray([]),
                     'rho': None, 'composition': None, 'sq_x': np.asarray([])}
        self.weights = {}
        self.c_dict = {}

    def set_atoms(self):

        _atom_list = self.data['composition'].keys()
        if self.structure_config_widget.polyatomic_gb.atom_list == _atom_list:
            pass

        else:
            self.structure_config_widget.polyatomic_gb.atom_list = _atom_list
            for _peak in self.structure_config_widget.polyatomic_gb.peak_dict.values():
                _peak.populate_atom_list(_atom_list)

    def plot_data(self, _reset_view=1):
        self.update_plot_data()
        self.structure_plot_widget.update_plot_windows(self.data, _reset_view)

    def update_plot_data(self):
        self.data['r0'] = self.structure_config_widget.monatomic_gb.r0_input.value()
        self.data['rpmax'] = self.structure_config_widget.monatomic_gb.rpmax_input.value()
        self.data['rmax'] = self.structure_config_widget.monatomic_gb.rmax_input.value()
        self.data['rmin'] = self.structure_config_widget.monatomic_gb.rmin_input.value()

        if not self.data['rdf_y'].size:
            return

        self.structure_plot_widget.update_plots(self.data)
        self.create_int_limit_signals()

    def update_obj_fun(self):
        if self.active_func == 'rdf':
            self.data['obj_fun'] = self.data['rdf_y']
        elif self.active_func == 'tr':
            self.data['obj_fun'] = self.data['tr_y']

    def toggle_plot_view(self):
        # Change active view
        if self.structure_config_widget.plot_view_gb.rdf_btn.isChecked():
            self.active_func = 'rdf'
            # Set label on curve fitting plot
            self.structure_plot_widget.fit_plot.setLabel('left', text='RDF(r)')
        elif self.structure_config_widget.plot_view_gb.tr_btn.isChecked():
            self.active_func = 'tr'
            self.structure_plot_widget.fit_plot.setLabel('left', text='T(r)')
        self.update_obj_fun()

        # Toggle plots if monatomic gb selected
        if self.structure_config_widget.monatomic_gb.isChecked():
            if self.structure_config_widget.plot_view_gb.rdf_btn.isChecked():
                self.structure_plot_widget.fit_layout_widget.setVisible(False)
                self.structure_plot_widget.tr_int_layout_widget.setVisible(False)
                self.structure_plot_widget.rdf_int_layout_widget.setVisible(True)
            elif self.structure_config_widget.plot_view_gb.tr_btn.isChecked():
                self.structure_plot_widget.fit_layout_widget.setVisible(False)
                self.structure_plot_widget.rdf_int_layout_widget.setVisible(False)
                self.structure_plot_widget.tr_int_layout_widget.setVisible(True)

        elif self.structure_config_widget.polyatomic_gb.isChecked():
            self.structure_plot_widget.fit_layout_widget.setVisible(True)
            self.structure_plot_widget.tr_int_layout_widget.setVisible(False)
            self.structure_plot_widget.rdf_int_layout_widget.setVisible(False)
            self.make_peak_plots()


    def toggle_monatomic_gb(self):
        if self.structure_config_widget.monatomic_gb.isChecked():
            # Toggled to enable monatomic (integration) gb
            # Disable gaussian fitting gb
            self.structure_config_widget.polyatomic_gb.setChecked(False)
            self.toggle_plot_view()

    def toggle_polyatomic_gb(self):
        if self.structure_config_widget.polyatomic_gb.isChecked():
            # Toggled to enable polyatomic (gaussian fitting) gb
            # Disable monatomic (integration) gb
            self.structure_config_widget.monatomic_gb.setChecked(False)
            self.toggle_plot_view()

    def update_fit_limits_sb(self):
        _min, _max = self.structure_plot_widget.fit_limits.getRegion()
        self.structure_config_widget.polyatomic_gb.min_limit_input.setValue(_min)
        self.structure_config_widget.polyatomic_gb.max_limit_input.setValue(_max)

    def update_fit_limits_plot(self):
        _min = self.structure_config_widget.polyatomic_gb.min_limit_input.value()
        _max = self.structure_config_widget.polyatomic_gb.max_limit_input.value()
        self.structure_plot_widget.fit_limits.setRegion((_min, _max))

    def auto_integration_limits(self):
        if not self.data['rdf_y'].size:
            return
        _r_0, _rp_max, _r_max, _r_min = data_utils.find_integration_limits(self.data['rdf_x'], self.data['rdf_y'])
        self.structure_config_widget.monatomic_gb.r0_input.setValue(_r_0)
        self.structure_config_widget.monatomic_gb.rpmax_input.setValue(_rp_max)
        self.structure_config_widget.monatomic_gb.rmax_input.setValue(_r_max)
        self.structure_config_widget.monatomic_gb.rmin_input.setValue(_r_min)

    def calc_integrals(self):
        if not self.data['rdf_y'].size:
            return
        _Na, _Nb, _Nc = core.integrate_coordination_sphere(self.data['rdf_x'], self.data['rdf_y'],
                                                           r_0=self.data['r0'], rp_max=self.data['rpmax'],
                                                           r_max=self.data['rmax'], r_min=self.data['rmin'],
                                                           method=0)
        self.structure_config_widget.monatomic_gb.Na_output.setText('{0:.3f}'.format(_Na))
        self.structure_config_widget.monatomic_gb.Nb_output.setText('{0:.3f}'.format(_Nb))
        self.structure_config_widget.monatomic_gb.Nc_output.setText('{0:.3f}'.format(_Nc))
        self.data['N_a'] = _Na
        self.data['N_b'] = _Nb
        self.data['N_c'] = _Nc

    def set_weights(self):
        if self.xray_weight_mode == 1:
            self.weights, self.c_dict = core.calculate_weights(self.data['composition'], self.data['sq_x'])
        elif self.xray_weight_mode == 0:
            self.weights, self.c_dict = core.calculate_weights(self.data['composition'], [0])
        else:
            raise ValueError

    # TODO - Proper guess peak implement plain gauss!
    def guess_peak_params(self, _idx):
        _peak_dict = self.structure_config_widget.polyatomic_gb.peak_dict
        _peak_widget = _peak_dict[_idx]
        if self.data['rdf_y'].size:
            # When data is present, guess the peak parameters
            # Not yet implemented - currently just set r to middle of
            # fitting range
            _lb, _ub = self.structure_plot_widget.fit_limits.getRegion()
            _lb = round(_lb, 1)
            _ub = round(_ub, 1)
            _pos = round((_lb + _ub)/2, 1)
            _peak_widget.r_input.setText(str(_pos))
            _peak_widget.r_lb.setText(str(_lb))
            _peak_widget.r_ub.setText(str(_ub))

        else:
            # If not data present set default parameters here
            # Currently default parameters are set for N/s in groupbox
            # Here we just set r to 3.0 as a default
            _peak_widget.r_input.setText('3.0')
            _peak_widget.r_lb.setText('1.0')
            _peak_widget.r_ub.setText('4.0')

    def make_peak_params(self):
        # Clear peak_fit_dict each time
        self.peak_fit_dict = {}
        self.fit_result = None

        _peak_dict = self.structure_config_widget.polyatomic_gb.peak_dict
        self.peak_dict = _peak_dict
        _peak_idx_list = list(_peak_dict.keys())
        _peak_idx_remove = []
        for _peak_idx in _peak_idx_list:
            _peak_idx_str = str(_peak_idx)
            _peak_widget = _peak_dict[_peak_idx]
            _alpha = _peak_widget.alpha_input.currentText()
            _beta = _peak_widget.beta_input.currentText()

            try:
                _N = np.float(_peak_widget.N_input.text())
            except ValueError:
                _N = 0.0
            _N_lb = np.float(_peak_widget.N_lb.text())
            _N_ub = np.float(_peak_widget.N_ub.text())
            _N_refine = _peak_widget.N_refine.isChecked()

            try:
                _r = np.float(_peak_widget.r_input.text())
            except ValueError:
                _r = 0.0
            _r_lb = np.float(_peak_widget.r_lb.text())
            _r_ub = np.float(_peak_widget.r_ub.text())
            _r_refine = _peak_widget.r_refine.isChecked()

            try:
                _s = np.float(_peak_widget.s_input.text())
            except ValueError:
                _s = 0.0
            _s_lb = np.float(_peak_widget.s_lb.text())
            _s_ub = np.float(_peak_widget.s_ub.text())
            _s_refine = _peak_widget.s_refine.isChecked()

            if _peak_widget.skew_toggle.isChecked():
                try:
                    _xi = np.float(_peak_widget.xi_input.text())
                except ValueError:
                    _xi = 0.0
                _xi_lb = np.float(_peak_widget.xi_lb.text())
                _xi_ub = np.float(_peak_widget.xi_ub.text())
                _xi_refine = _peak_widget.xi_refine.isChecked()

            else:
                _xi = 0.0
                _xi_lb = 0.0
                _xi_ub = 1.0
                _xi_refine = False

            # Make peak name
            _name = '#' + _peak_idx_str + ' ' + _alpha + '-' + _beta

            try:
                # Make kwarg dict
                _W = self.weights[(_alpha,_beta)]
                _cb = self.c_dict[_beta]
                _kwargs = {'W'+_peak_idx_str: _W, 'c_b'+_peak_idx_str: _cb}
                # Make parameter object
                _params = lmfit.Parameters()
                # Form for lmfit parameters
                # (NAME VALUE VARY MIN  MAX  EXPR  BRUTE_STEP)
                _params.add_many(
                    ('N'+_peak_idx_str, _N, _N_refine, _N_lb, _N_ub, None, None),
                    ('r'+_peak_idx_str, _r, _r_refine, _r_lb, _r_ub, None, None),
                    ('s'+_peak_idx_str, _s, _s_refine, _s_lb, _s_ub, None, None),
                    ('xi'+_peak_idx_str, _xi, _xi_refine, _xi_lb, _xi_ub, None, None))
                # Store name, params, kwargs in peak_fit_dict
                self.peak_fit_dict['name'+_peak_idx_str] = _name
                self.peak_fit_dict['kwargs'+_peak_idx_str] = _kwargs
                self.peak_fit_dict['params'+_peak_idx_str] = _params

            except KeyError:
                # If alpha/beta not set remove peak from the list
                _peak_idx_remove.append(_peak_idx)

        # Remove peaks with alpha-beta not set
        for _peak_to_remove in _peak_idx_remove:
            _peak_idx_list.remove(_peak_to_remove)
        # Store peak_idx_list in peak_fit_dict
        self.peak_fit_dict['peak_idx_list'] = _peak_idx_list

        # Combine param objects
        _obj_params = lmfit.Parameters()
        for _peak_idx in _peak_idx_list:
            _peak_idx_str = str(_peak_idx)
            _peak_params = self.peak_fit_dict['params'+_peak_idx_str]
            for _param in list(_peak_params):
                _obj_params.add(_peak_params[_param])
        # Store combined parameters for objective function
        self.peak_fit_dict['obj_params'] = _obj_params

        # Combine kwarg dicts for objective function
        _obj_kwargs = {}
        for _peak_idx in _peak_idx_list:
            _peak_idx_str = str(_peak_idx)
            _obj_kwargs.update(self.peak_fit_dict['kwargs'+_peak_idx_str])
            # Stor peak idx (as peak_idx_list) in each kwarg dict
            self.peak_fit_dict['kwargs'+_peak_idx_str]['peak_idx_list'] = [_peak_idx]
            self.peak_fit_dict['kwargs'+_peak_idx_str]['active_func'] = self.active_func

        # Store combined kwargs in peak_fit_dict
        self.peak_fit_dict['obj_kwargs'] = _obj_kwargs
        # Store peak_idx_list in objective kwargs also
        self.peak_fit_dict['obj_kwargs']['peak_idx_list'] = _peak_idx_list
        # And the active obj func
        self.peak_fit_dict['obj_kwargs']['active_func'] = self.active_func

        # Cut data at the fit range
        _fit_min = self.structure_config_widget.polyatomic_gb.min_limit_input.value()
        _fit_max = self.structure_config_widget.polyatomic_gb.max_limit_input.value()
        _fit_range = np.where((self.data['rdf_x'] >= _fit_min) & (self.data['rdf_x'] <= _fit_max))
        self.data['fit_r'] = self.data['rdf_x'][_fit_range]
        self.data['fit_y'] = self.data['obj_fun'][_fit_range]

    def on_click_fit_peaks(self):

        self.make_peak_params()
        # Exit if no peaks present
        if not self.peak_fit_dict['peak_idx_list']:
            return

        # Double check there are no NaNs in data-set - can occur if x-lim is 0
        if np.isnan(self.data['fit_y']).any():
            self.data['fit_y'] = data_utils.interp_nan(self.data['fit_y'])

        self.fit_result = lmfit.minimize(peak_fit.gauss_obj_func,
                                         self.peak_fit_dict['obj_params'],
                                         args=(self.data['fit_r'], self.data['fit_y']),
                                         kws=self.peak_fit_dict['obj_kwargs'])

        # Save peak fit results to file
        # Make log of peak IDs/atom-pairs
        _peak_log = (f'Peak IDs\n{"-"*8}\n')
        for _peak_idx in self.peak_fit_dict['peak_idx_list']:
            _peak_id_str = f'{_peak_idx} :  {self.peak_fit_dict["name"+str(_peak_idx)]}\n'
            _peak_log += _peak_id_str
        # Make log string
        log_string = (
            f'refinement_log (fit peaks)\n'
            f'{__appname__} v{__version__}\n\n'
            f'{"-"*25}\n'
            f'Data File : {self.base_filename}{self.filename_ext}\n'
            f'Composition [Element: (Z, Charge, n)]: {self.data["composition"]}\n'
            f'{_peak_log}'
            f'{"*"*25}\n'
            f'Using lmfit v{lmfit.__version__}\n'
            f'Peak fit success?: {self.fit_result.success} | {self.fit_result.message}\n'
            f'{lmfit.fit_report(self.fit_result)}\n'
        )
        # Append to log file or overwrite?
        if self.append_log_mode:
            _log_file = open(os.path.join(os.path.dirname(self.base_filename), 'refinement.log'), 'ab')
            # Add timestamp to top of log_string
            log_string = (f'{"#"*30}\n'
                          f'{datetime.datetime.now()}\n'
                          f'{"#"*30}\n' +
                          log_string
                          )
            np.savetxt(_log_file, [log_string], fmt='%s')
            _log_file.close()
        else:
            _log_file = self.base_filename + '_refinement.log'
            np.savetxt(_log_file, [log_string], fmt='%s')

        # Set refined peak parameters
        self.set_peak_params()

    def set_peak_params(self):

        # Get parameter object
        _params = self.fit_result.params
        # can loop through peak list
        _peak_idx_list = self.peak_fit_dict['peak_idx_list']
        for _peak_idx in _peak_idx_list:
            _peak_idx_str = str(_peak_idx)
            _peak_widget = self.peak_dict[_peak_idx]
            # Reset alpha and beta
            _name = self.peak_fit_dict['name'+_peak_idx_str]
            _alpha, _beta = _name.split(' ')[-1].split('-')
            _peak_widget.alpha_input.setCurrentText(_alpha)
            _peak_widget.beta_input.setCurrentText(_beta)
            # Set N
            _N_param = _params['N'+_peak_idx_str]
            _peak_widget.N_input.setText('{0:.2f}'.format(_N_param.value))
            _peak_widget.N_lb.setText('{0:.2f}'.format(_N_param.min))
            _peak_widget.N_ub.setText('{0:.2f}'.format(_N_param.max))
            _peak_widget.N_refine.setChecked(_N_param.vary)

            # Set r
            _r_param = _params['r'+_peak_idx_str]
            _peak_widget.r_input.setText('{0:.2f}'.format(_r_param.value))
            _peak_widget.r_lb.setText('{0:.2f}'.format(_r_param.min))
            _peak_widget.r_ub.setText('{0:.2f}'.format(_r_param.max))
            _peak_widget.r_refine.setChecked(_r_param.vary)

            # Set sigma
            _s_param = _params['s'+_peak_idx_str]
            _peak_widget.s_input.setText('{0:.2f}'.format(_s_param.value))
            _peak_widget.s_lb.setText('{0:.2f}'.format(_s_param.min))
            _peak_widget.s_ub.setText('{0:.2f}'.format(_s_param.max))
            _peak_widget.s_refine.setChecked(_s_param.vary)

            # Set xi (skewness) only if in use
            if _peak_widget.skew_toggle.isChecked():
                _xi_param = _params['xi'+_peak_idx_str]
                _peak_widget.xi_input.setText('{0:.2f}'.format(_xi_param.value))
                _peak_widget.xi_lb.setText('{0:.2f}'.format(_xi_param.min))
                _peak_widget.xi_ub.setText('{0:.2f}'.format(_xi_param.max))
                _peak_widget.xi_refine.setChecked(_xi_param.vary)

        # Clear the peak fit dict
        self.peak_fit_dict = {}
        # Clear fitting results
        self.fit_result = None

    def make_peak_plots(self):
        # Make peak_fit_dict
        self.make_peak_params()
        # Do not plot if no peaks present
        if not self.peak_fit_dict['peak_idx_list']:
            self.structure_plot_widget.clear_gauss_curves()
            self.data['gauss_model'] = np.asarray([])
            self.data['gauss_peaks'] = []
            self.plot_data(_reset_view=0)
            return
        # Evaluate objective function over whole r range
        self.data['gauss_model'] = peak_fit.gauss_obj_func(self.peak_fit_dict['obj_params'],
                                                           self.data['rdf_x'], data=None,
                                                           **self.peak_fit_dict['obj_kwargs'])

        # Calculate residuals from objective function over min < r < max
        self.data['gauss_residuals'] = peak_fit.gauss_obj_func(self.peak_fit_dict['obj_params'],
                                                               self.data['fit_r'], data=self.data['fit_y'],
                                                               **self.peak_fit_dict['obj_kwargs'])
        # Calculate residuals from objective function over total r range
        self.data['gauss_residuals_full'] = peak_fit.gauss_obj_func(
                                                self.peak_fit_dict['obj_params'],
                                                self.data['rdf_x'], data=self.data['obj_fun'],
                                                **self.peak_fit_dict['obj_kwargs'])

        # Evaluate each individual peak over r range
        # List to store each individual peak
        self.data['gauss_peaks'] = []
        self.data['gauss_peaks_names'] = []
        for _peak_idx in self.peak_fit_dict['peak_idx_list']:
            _peak_idx_str = str(_peak_idx)
            _peak_model = peak_fit.gauss_obj_func(self.peak_fit_dict['params'+_peak_idx_str],
                                                  self.data['rdf_x'], data=None,
                                                  **self.peak_fit_dict['kwargs'+_peak_idx_str])
            self.data['gauss_peaks'].append(_peak_model)
            self.data['gauss_peaks_names'].append(self.peak_fit_dict['name'+_peak_idx_str])

        # Call update_plots
        self.structure_plot_widget.update_plots(self.data)
        self.structure_plot_widget.set_res_window(self.data)


class StructureConfigWidget(QWidget):

    def __init__(self):
        super(StructureConfigWidget, self).__init__()

        self.vlayout = QVBoxLayout()
        self.vlayout.setContentsMargins(0, 0, 5, 0)
        self.vlayout.setSpacing(10)

        self.plot_view_gb = PlotViewGroupBox()
        self.monatomic_gb = MonatomicGroupBox()
        self.polyatomic_gb = PolyatomicGroupBox()

        self.vlayout.addWidget(self.plot_view_gb, 1)
        self.vlayout.addWidget(self.monatomic_gb, 1)
        self.vlayout.addWidget(self.polyatomic_gb, 1)

        # Dummy widget to scale vertically
        self.vlayout.addWidget(QWidget(), 3)

        self.setLayout(self.vlayout)


class PlotViewGroupBox(QGroupBox):

    def __init__(self, *args):
        super(PlotViewGroupBox, self).__init__(*args)
        self.setTitle('Plot View')
        self.setAlignment(Qt.AlignLeft)
        self.setStyleSheet('GroupBox::title{subcontrol-origin: margin; subcontrol-position: top left;}')

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

    def create_widgets(self):
        self.plot_view_label = QLabel('Select function: ')
        self.rdf_btn = QRadioButton('RDF(r)')
        self.tr_btn = QRadioButton('T(r)')

    def style_widgets(self):
        self.rdf_btn.setToolTip('RDF(r) = 4πρr<sup>2</sup>g(r)')
        self.tr_btn.setToolTip('T(r) = RDF(r) / r')
        self.rdf_btn.setChecked(True)

    def create_layout(self):
        self.main_layout = QVBoxLayout()
        self.main_layout.setContentsMargins(20, 1, 20, 5)
        self.main_layout.setSpacing(1)

        self.hbtn_layout = QGridLayout()
        self.hbtn_layout.setContentsMargins(0, 0, 0, 0)
        self.hbtn_layout.addWidget(self.rdf_btn, 0, 0)
        self.hbtn_layout.addWidget(self.tr_btn, 0, 1)

        self.main_layout.addWidget(self.plot_view_label)
        self.main_layout.addLayout(self.hbtn_layout)
        self.setLayout(self.main_layout)

    def create_signals(self):
        pass


class MonatomicGroupBox(QGroupBox):

    def __init__(self, *args):
        super(MonatomicGroupBox, self).__init__(*args)
        self.setTitle('Integrate N-1 (Monatomic)')
        self.setAlignment(Qt.AlignLeft)
        self.setStyleSheet('GroupBox::title{subcontrol-origin: margin; subcontrol-position: top left;}')
        self.setCheckable(True)
        self.setChecked(True)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

    def create_widgets(self):

        self.r0_label = QLabel('r<sub>0</sub>')
        self.r0_input = QDoubleSpinBox()

        self.rpmax_label = QLabel('r\'<sub>max</sub>')
        self.rpmax_input = QDoubleSpinBox()

        self.rmax_label = QLabel('r<sub>max</sub>')
        self.rmax_input = QDoubleSpinBox()

        self.rmin_label = QLabel('r<sub>min</sub>')
        self.rmin_input = QDoubleSpinBox()

        self.find_limits_btn = QPushButton('Auto-Refine Int. Limits')

        self.calc_N_btn = QPushButton('Calc Coordination Number')

        self.Na_output_label = QLabel('N<sub>a</sub> : ')
        self.Na_output = QLineEdit()

        self.Nb_output_label = QLabel('N<sub>b</sub> : ')
        self.Nb_output = QLineEdit()

        self.Nc_output_label = QLabel('N<sub>c</sub> : ')
        self.Nc_output = QLineEdit()

    def style_widgets(self):

        self.r0_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.rpmax_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.rmax_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.rmin_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        self.Na_output_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.Nb_output_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.Nc_output_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        # May need to change max widths
        self.r0_input.setMaximumWidth(82)
        self.rpmax_input.setMaximumWidth(82)
        self.rmax_input.setMaximumWidth(82)
        self.rmin_input.setMaximumWidth(82)

        self.Na_output.setMaximumWidth(100)
        self.Nb_output.setMaximumWidth(100)
        self.Nc_output.setMaximumWidth(100)

        # Set results as read-only
        self.Na_output.setReadOnly(True)
        self.Nb_output.setReadOnly(True)
        self.Nc_output.setReadOnly(True)

        # Set input spinbox settings
        # Might need to set minimum width also to stop scale miss
        self.r0_input.setSingleStep(0.01)
        self.r0_input.setDecimals(3)
        self.r0_input.setAlignment(Qt.AlignRight)
        self.rpmax_input.setSingleStep(0.01)
        self.rpmax_input.setDecimals(3)
        self.rpmax_input.setAlignment(Qt.AlignRight)
        self.rmax_input.setSingleStep(0.01)
        self.rmax_input.setDecimals(3)
        self.rmax_input.setAlignment(Qt.AlignRight)
        self.rmin_input.setSingleStep(0.01)
        self.rmin_input.setDecimals(3)
        self.rmin_input.setAlignment(Qt.AlignRight)

        # Set tooltips
        self.r0_label.setToolTip('Leading edge of first peak in RDF(r)')
        self.r0_input.setToolTip('Leading edge of first peak in RDF(r)')

        self.rmin_input.setToolTip('First minimum after first peak in RDF(r)')
        self.rmin_label.setToolTip('First minimum after first peak in RDF(r)')

        self.rpmax_label.setToolTip('Centre of first peak in T(r)')
        self.rpmax_input.setToolTip('Centre of first peak in T(r)')

        self.rmax_label.setToolTip('Centre of first peak in RDF(r)')
        self.rmax_input.setToolTip('Centre of first peak in RDF(r)')

        self.Na_output_label.setToolTip('N<sub>a</sub> = 2 ∫ 4πρr[rg(r)]<sub>sym</sub> dr | r = r<sub>0</sub> to r = r\'<sub>max</sub>')
        self.Na_output.setToolTip('N<sub>a</sub> = 2 ∫ 4πρr[rg(r)]<sub>sym</sub> dr | r = r<sub>0</sub> to r = r\'<sub>max</sub>')

        self.Nb_output_label.setToolTip('N<sub>b</sub> = 2 ∫ 4πρ[r<sup>2</sup>g(r)]<sub>sym</sub> dr | r = r<sub>0</sub> to r = r<sub>max</sub>')
        self.Nb_output.setToolTip('N<sub>b</sub> = 2 ∫ 4πρ[r<sup>2</sup>g(r)]<sub>sym</sub> dr | r = r<sub>0</sub> to r = r<sub>max</sub>')

        self.Nc_output_label.setToolTip('N<sub>c</sub> = ∫ 4πρr<sup>2</sup>g(r) dr | r = r<sub>0</sub> to r = r<sub>min</sub>')
        self.test = self.Nc_output.setToolTip('N<sub>c</sub> = ∫ 4πρr<sup>2</sup>g(r) dr | r = r<sub>0</sub> to r = r<sub>min</sub>')

        self.setStyleSheet(" QToolTip{ font: 11pt;}")
        # Set ToolTip style sheets separately
        self.Na_output_label.setStyleSheet(" QToolTip{ min-width: 480px; font: 18pt;}")
        self.Na_output.setStyleSheet(" QToolTip{ min-width: 480px; font: 18pt;}")
        self.Nb_output_label.setStyleSheet(" QToolTip{ min-width: 475px; font: 18pt;}")
        self.Nb_output.setStyleSheet(" QToolTip{ min-width: 475px; font: 18pt;}")
        self.Nc_output_label.setStyleSheet(" QToolTip{ min-width: 400px; font: 18pt;}")
        self.Nc_output.setStyleSheet(" QToolTip{ min-width: 400px; font: 18pt;}")  

    def create_layout(self):
        self.outer_layout = QVBoxLayout()
        self.outer_layout.setContentsMargins(25, 10, 25, 7)
        self.outer_layout.setSpacing(30)

        self.int_limits_layout = QVBoxLayout()
        self.int_limits_layout.setContentsMargins(0, 0, 0, 0)
        self.int_limits_layout.setSpacing(15)

        self.int_results_layout = QVBoxLayout()
        self.int_results_layout.setContentsMargins(0, 0, 0, 0)
        self.int_results_layout.setSpacing(15)

        # Grid for integration limits
        self.int_limits_grid = QGridLayout()
        self.int_limits_grid.setContentsMargins(0, 0, 0, 0)
        self.int_limits_grid.setSpacing(15)
        self.int_limits_grid.addWidget(self.r0_label, 0, 0)
        self.int_limits_grid.addWidget(self.r0_input, 0, 1)
        self.int_limits_grid.addWidget(self.rmin_label, 0, 2)
        self.int_limits_grid.addWidget(self.rmin_input, 0, 3)
        self.int_limits_grid.addWidget(self.rpmax_label, 1, 0)
        self.int_limits_grid.addWidget(self.rpmax_input, 1, 1)
        self.int_limits_grid.addWidget(self.rmax_label, 1, 2)
        self.int_limits_grid.addWidget(self.rmax_input, 1, 3)

        # Grid for N1 results
        self.int_results_grid = QGridLayout()
        self.int_results_grid.setContentsMargins(0, 0, 0, 0)
        self.int_results_grid.setSpacing(15)
        self.int_results_grid.addWidget(self.Na_output_label, 0, 0)
        self.int_results_grid.addWidget(self.Na_output, 0, 1)
        self.int_results_grid.addWidget(self.Nb_output_label, 1, 0)
        self.int_results_grid.addWidget(self.Nb_output, 1, 1)
        self.int_results_grid.addWidget(self.Nc_output_label, 2, 0)
        self.int_results_grid.addWidget(self.Nc_output, 2, 1)

        self.int_limits_layout.addWidget(self.find_limits_btn)
        self.int_limits_layout.addLayout(self.int_limits_grid)

        self.int_results_layout.addWidget(self.calc_N_btn)
        self.int_results_layout.addLayout(self.int_results_grid)

        self.outer_layout.addLayout(self.int_limits_layout)
        self.outer_layout.addLayout(self.int_results_layout)

        self.setLayout(self.outer_layout)

    def create_signals(self):
        pass


class PolyatomicGroupBox(QGroupBox):

    # Create custom signals emmitted when peaks added/changed
    guess_peak = pyqtSignal(int)
    # Relay signal from GaussianPeakGroupBox
    gauss_params_changed_relay = pyqtSignal()

    def __init__(self, *args):
        super(PolyatomicGroupBox, self).__init__(*args)
        self.setTitle('Fit Gaussians (Polyatomic)')
        self.setAlignment(Qt.AlignLeft)
        self.setStyleSheet('GroupBox::title{subcontrol-origin: margin; subcontrol-position: top left;}')
        self.setCheckable(True)
        self.setChecked(False)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()

        self.atom_list = []
        self.init_peaks()

        self.create_signals()

    def create_widgets(self):

        self.fit_limit_label = QLabel('Region to fit: ')
        self.min_limit_label = QLabel('min: ')
        self.max_limit_label = QLabel('max: ')

        self.min_limit_input = QDoubleSpinBox()
        self.max_limit_input = QDoubleSpinBox()

        self.fit_peaks_btn = QPushButton('Fit')

        self.add_peak_btn = QPushButton('Add Peak')

    def style_widgets(self):

        self.fit_limit_label.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        self.min_limit_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.max_limit_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        self.min_limit_input.setMaximumWidth(82)
        self.max_limit_input.setMaximumWidth(82)

        self.min_limit_input.setSingleStep(0.1)
        self.min_limit_input.setDecimals(2)
        self.min_limit_input.setAlignment(Qt.AlignRight)

        self.max_limit_input.setSingleStep(0.1)
        self.max_limit_input.setDecimals(2)
        self.max_limit_input.setAlignment(Qt.AlignRight)

    def create_layout(self):
        self.outer_layout = QVBoxLayout()
        self.outer_layout.setContentsMargins(25, 10, 25, 7)
        self.outer_layout.setSpacing(30)

        self.fit_settings_grid = QGridLayout()
        self.fit_settings_grid.setContentsMargins(0, 0, 0, 0)
        self.fit_settings_grid.setSpacing(15)

        self.fit_settings_grid.addWidget(self.fit_limit_label, 0, 0)
        self.fit_settings_grid.addWidget(self.min_limit_label, 0, 1)
        self.fit_settings_grid.addWidget(self.min_limit_input, 0, 2)
        self.fit_settings_grid.addWidget(self.max_limit_label, 1, 1)
        self.fit_settings_grid.addWidget(self.max_limit_input, 1, 2)

        self.outer_layout.addLayout(self.fit_settings_grid)
        self.outer_layout.addWidget(self.fit_peaks_btn)
        self.outer_layout.addWidget(self.add_peak_btn)

        self.setLayout(self.outer_layout)

    def init_peaks(self):
        self.peak_count = 1
        self.peak_dict = {}

    def create_signals(self):

        self.add_peak_btn.clicked.connect(self.add_peak)

        self.del_peak_signalMapper = QSignalMapper()
        self.del_peak_signalMapper.mapped[int].connect(self.del_peak)

        self.rename_peak_signalMapper = QSignalMapper()
        self.rename_peak_signalMapper.mapped[int].connect(self.rename_peak)

        self.min_limit_input.valueChanged.connect(self.gauss_params_changed_relay)
        self.max_limit_input.valueChanged.connect(self.gauss_params_changed_relay)

    def add_peak(self):

        if len(self.peak_dict) == 0:
            self.init_peaks()

        _new_peak = GaussianPeakGroupBox()
        _new_peak._unique_idx = self.peak_count
        _new_peak.title.setText(_new_peak.title.text() + ' #' + str(_new_peak._unique_idx) + ' ')
        _new_peak.populate_atom_list(self.atom_list)

        self.peak_dict[_new_peak._unique_idx] = _new_peak
        self.outer_layout.addWidget(_new_peak)

        # Link peak delete button to signal mapper
        self.del_peak_signalMapper.setMapping(self.peak_dict[_new_peak._unique_idx].del_peak_btn, self.peak_dict[_new_peak._unique_idx]._unique_idx)
        self.peak_dict[_new_peak._unique_idx].del_peak_btn.clicked.connect(self.del_peak_signalMapper.map)

        # Link atom inputs to name of peak
        self.rename_peak_signalMapper.setMapping(self.peak_dict[_new_peak._unique_idx].alpha_input, self.peak_dict[_new_peak._unique_idx]._unique_idx)
        self.peak_dict[_new_peak._unique_idx].alpha_input.currentTextChanged.connect(self.rename_peak_signalMapper.map)

        self.rename_peak_signalMapper.setMapping(self.peak_dict[_new_peak._unique_idx].beta_input, self.peak_dict[_new_peak._unique_idx]._unique_idx)
        self.peak_dict[_new_peak._unique_idx].beta_input.currentTextChanged.connect(self.rename_peak_signalMapper.map)

        self.peak_count += 1

        # Re-emit parameter changed signal
        self.peak_dict[_new_peak._unique_idx].gauss_params_changed.connect(self.gauss_params_changed_relay)

        # Emit signal to guess peak parameters of new peak
        self.guess_peak.emit(_new_peak._unique_idx)

        del _new_peak

    def del_peak(self, _peak_index):

        # Remove from layout
        self.outer_layout.removeWidget(self.peak_dict[_peak_index])
        # Delete C/C++ object
        self.peak_dict[_peak_index].deleteLater()
        # Delete Py object
        self.peak_dict[_peak_index] = None
        # Remove peak entry from dict
        del self.peak_dict[_peak_index]
        del _peak_index
        # Update the plots
        self.gauss_params_changed_relay.emit()

    def rename_peak(self, _peak_index):

        _target_peak = self.peak_dict[_peak_index]
        _alpha = _target_peak.alpha_input.currentText()
        _beta = _target_peak.beta_input.currentText()
        _current_title = _target_peak.title.text().split('(')[0]
        if (_alpha == '') & (_beta == ''):
            _target_peak.title.setText(_current_title)
        else:
            _title = _current_title + '(' + _alpha + '-' + _beta + ')'
            _target_peak.title.setText(_title)


class GaussianPeakGroupBox(QFrame):

    # Create signal emmitted on parameter change
    gauss_params_changed = pyqtSignal()

    def __init__(self, *args):
        super(GaussianPeakGroupBox, self).__init__(*args)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

    def create_widgets(self):

        self.title = QLabel('Gaussian Peak')

        self.hline = QFrame()
        self.hline.setFrameShape(QFrame.HLine)
        self.hline.setFrameShadow(QFrame.Sunken)
        self.hline.setObjectName('hline')

        self.del_peak_btn = QPushButton()

        self.alpha_label = QLabel('alpha')
        self.beta_label = QLabel('beta')

        self.alpha_input = QComboBox()
        self.alpha_input.setCurrentIndex(-1)

        self.beta_input = QComboBox()
        self.beta_input.setCurrentIndex(-1)

        self.refine_label = QLabel('Refine?')
        self.lb_label = QLabel('min')
        self.ub_label = QLabel('max')

        self.N_label = QLabel('N: ')
        self.N_input = QLineEdit('2.0')
        self.N_lb = QLineEdit('0')
        self.N_ub = QLineEdit('10.0')
        self.N_refine = QCheckBox()

        self.r_label = QLabel('r: ')
        self.r_input = QLineEdit('2.0')
        self.r_lb = QLineEdit('1.0')
        self.r_ub = QLineEdit('3.0')
        self.r_refine = QCheckBox()

        self.s_label = QLabel('σ: ')
        self.s_input = QLineEdit('0.1')
        self.s_lb = QLineEdit('0')
        self.s_ub = QLineEdit('1.0')
        self.s_refine = QCheckBox()

        self.skew_toggle = QCheckBox('Skew?: ')
        self.xi_label = QLabel('ξ: ')
        self.xi_input = QLineEdit('0.01')
        self.xi_lb = QLineEdit('0')
        self.xi_ub = QLineEdit('2.0')
        self.xi_refine = QCheckBox()

    def style_widgets(self):

        self.title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)

        self.del_peak_btn.setIcon(self.style().standardIcon(getattr(QStyle, 'SP_TitleBarCloseButton')))
        self.del_peak_btn.setMaximumWidth(25)
        self.del_peak_btn.setFlat(True)

        self.alpha_label.setAlignment(Qt.AlignVCenter | Qt.AlignCenter)
        self.beta_label.setAlignment(Qt.AlignVCenter | Qt.AlignCenter)

        self.refine_label.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)

        self.lb_label.setAlignment(Qt.AlignVCenter | Qt.AlignCenter)
        self.ub_label.setAlignment(Qt.AlignVCenter | Qt.AlignCenter)

        self.N_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.r_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.s_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)
        self.xi_label.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        self.skew_toggle.setLayoutDirection(Qt.RightToLeft)
        #self.skew_toggle.text.setAlignment(Qt.AlignVCenter)

        self.N_input.setMaximumWidth(100)
        self.r_input.setMaximumWidth(100)
        self.s_input.setMaximumWidth(100)
        self.xi_input.setMaximumWidth(100)

        self.N_lb.setMaximumWidth(100)
        self.N_ub.setMaximumWidth(100)
        self.r_lb.setMaximumWidth(100)
        self.r_ub.setMaximumWidth(100)
        self.s_lb.setMaximumWidth(100)
        self.s_ub.setMaximumWidth(100)
        self.xi_lb.setMaximumWidth(100)
        self.xi_ub.setMaximumWidth(100)

        # Set validators
        self.N_input.setValidator(QDoubleValidator())
        self.r_input.setValidator(QDoubleValidator())
        self.s_input.setValidator(QDoubleValidator())
        self.xi_input.setValidator(QDoubleValidator())
        self.N_lb.setValidator(QDoubleValidator())
        self.N_ub.setValidator(QDoubleValidator())
        self.r_lb.setValidator(QDoubleValidator())
        self.r_ub.setValidator(QDoubleValidator())
        self.s_lb.setValidator(QDoubleValidator())
        self.s_ub.setValidator(QDoubleValidator())
        self.xi_lb.setValidator(QDoubleValidator())
        self.xi_ub.setValidator(QDoubleValidator())

        self.N_refine.setChecked(True)
        self.r_refine.setChecked(True)
        self.s_refine.setChecked(True)
        self.xi_refine.setChecked(True)

        self.skew_toggle.setChecked(False)
        self.toggle_snd()

    def create_layout(self):

        self.vlayout = QVBoxLayout()
        self.vlayout.setContentsMargins(0, 5, 0, 0)
        self.vlayout.setSpacing(15)

        self.title_layout = QVBoxLayout()
        self.title_layout.setContentsMargins(0, 0, 0, 0)
        self.title_layout.setSpacing(5)
        self.title_inner_layout = QHBoxLayout()
        self.title_inner_layout.setContentsMargins(0, 0, 0, 0)
        self.title_inner_layout.setSpacing(0)

        self.atoms_grid_layout = QGridLayout()
        self.atoms_grid_layout.setContentsMargins(0, 0, 0, 0)
        self.atoms_grid_layout.setSpacing(10)

        self.title_inner_layout.addWidget(self.title)
        self.title_inner_layout.addWidget(self.del_peak_btn)
        self.title_layout.addLayout(self.title_inner_layout)
        self.title_layout.addWidget(self.hline)

        self.atoms_grid_layout.addWidget(self.alpha_label, 0, 0)
        self.atoms_grid_layout.addWidget(self.beta_label, 0, 1)
        self.atoms_grid_layout.addWidget(self.alpha_input, 1, 0)
        self.atoms_grid_layout.addWidget(self.beta_input, 1, 1)

        self.params_grid_layout = QGridLayout()
        self.params_grid_layout.setContentsMargins(0, 0, 0, 0)
        self.params_grid_layout.setSpacing(10)

        self.params_grid_layout.addWidget(self.lb_label, 0, 2)
        self.params_grid_layout.addWidget(self.ub_label, 0, 3)

        self.params_grid_layout.addWidget(self.refine_label, 0, 4)

        self.params_grid_layout.addWidget(self.N_label, 1, 0)
        self.params_grid_layout.addWidget(self.N_input, 1, 1)
        self.params_grid_layout.addWidget(self.N_lb, 1, 2)
        self.params_grid_layout.addWidget(self.N_ub, 1, 3)
        self.params_grid_layout.addWidget(self.N_refine, 1, 4)

        self.params_grid_layout.addWidget(self.r_label, 2, 0)
        self.params_grid_layout.addWidget(self.r_input, 2, 1)
        self.params_grid_layout.addWidget(self.r_lb, 2, 2)
        self.params_grid_layout.addWidget(self.r_ub, 2, 3)
        self.params_grid_layout.addWidget(self.r_refine, 2, 4)

        self.params_grid_layout.addWidget(self.s_label, 3, 0)
        self.params_grid_layout.addWidget(self.s_input, 3, 1)
        self.params_grid_layout.addWidget(self.s_lb, 3, 2)
        self.params_grid_layout.addWidget(self.s_ub, 3, 3)
        self.params_grid_layout.addWidget(self.s_refine, 3, 4)

        self.params_grid_layout.addWidget(self.skew_toggle, 4, 0, 1, 2)

        self.params_grid_layout.addWidget(self.xi_label, 5, 0)
        self.params_grid_layout.addWidget(self.xi_input, 5, 1)
        self.params_grid_layout.addWidget(self.xi_lb, 5, 2)
        self.params_grid_layout.addWidget(self.xi_ub, 5, 3)
        self.params_grid_layout.addWidget(self.xi_refine, 5, 4)

        self.vlayout.addLayout(self.title_layout)
        self.vlayout.addLayout(self.atoms_grid_layout)
        self.vlayout.addLayout(self.params_grid_layout)

        self.setLayout(self.vlayout)

    def create_signals(self):
        self.alpha_input.currentIndexChanged.connect(self.emit_param_change)
        self.beta_input.currentIndexChanged.connect(self.emit_param_change)
        self.N_input.textChanged.connect(self.emit_param_change)
        self.r_input.textChanged.connect(self.emit_param_change)
        self.s_input.textChanged.connect(self.emit_param_change)
        self.xi_input.textChanged.connect(self.emit_param_change)
        self.skew_toggle.stateChanged.connect(self.toggle_snd)

    def emit_param_change(self):
        self.gauss_params_changed.emit()

    def populate_atom_list(self, _atom_list):
        self.alpha_input.clear()
        self.beta_input.clear()
        self.alpha_input.addItems(_atom_list)
        self.beta_input.addItems(_atom_list)
        self.alpha_input.setCurrentIndex(-1)
        self.beta_input.setCurrentIndex(-1)

    def toggle_snd(self):
        if self.skew_toggle.isChecked():
            self.xi_label.setEnabled(True)
            self.xi_input.setEnabled(True)
            self.xi_lb.setEnabled(True)
            self.xi_ub.setEnabled(True)
            self.xi_refine.setEnabled(True)
        else:
            self.xi_label.setEnabled(False)
            self.xi_input.setEnabled(False)
            self.xi_lb.setEnabled(False)
            self.xi_ub.setEnabled(False)
            self.xi_refine.setEnabled(False)
        self.emit_param_change()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
LiquidDiffract core functions module

Python implementation of iterative method for liquid structure factor
normalisation based on Eggert et al., 2002.
Optimises the normalisation factor, alpha, and allows refinement of sample density, rho.

Provides functions for:

  * Calculation of liquid structure factor, S(Q)
  * Fourier transformation of S(Q)/g(r)
  * Calculation of pair and radial distribution functions, g(r) and RDF(r)
  * Calculation of atomic form factors, effective electronic form factor, and scattering intensities
  * Utility functions for density and molecular mass conversions/automatic calculation

See <https://github.com/bjheinen/LiquidDiffract#using-liquiddiffract-core-library> for more details

"""
__author__ = 'Benedict J. Heinen'
__copyright__ = 'Copyright 2018-2021, Benedict J. Heinen'
__email__ = 'benedict.heinen@gmail.com'
__license__ = 'GNU GPL v3'

from functools import lru_cache
from itertools import product as cartesian_product
import numpy as np
from scipy.integrate import simps, quadrature
import scipy.interpolate
import scipy.fftpack
# importlib.resources only available in python>=3.7
try:
    from importlib import resources as importlib_resources
except ImportError:
    import importlib_resources

# Get version number from version.py
from LiquidDiffract.version import __appname__, __version__
from LiquidDiffract.core.data_utils import data_cache
from LiquidDiffract.core.data_utils import smooth_data


def calc_mol_mass(composition):
    '''
    Calculates the molecular mass for a given composition and returns
    the average molecular mass for one atom

    composition is dictionary in the form: (Z, charge, n)
    where n is number of atoms in formula unit
    '''
    with importlib_resources.open_binary('LiquidDiffract.resources', 'mass_data.npy') as fp:
        mass_dict = np.load(fp, allow_pickle=True).item()
    mol_mass = np.sum([mass_dict[element]*(composition[element][2]) for element in composition])
    return mol_mass


def conv_density(rho, composition):
    '''
    Converts atomic number density - atoms/angstrom^3
    to mass density in g/cm^3
    '''
    # Calculate molecular mass g/mole
    mol_mass = calc_mol_mass(composition)
    # Calculate number of atoms and scale mol_mass
    N = np.sum(composition[el][2] for el in composition)
    # Calculate atoms/cm3 > then moles/cm3
    with np.errstate(divide='ignore', invalid='ignore'):
        mass_density = (rho * 10 / 6.0221408) * (mol_mass/N)
    return mass_density


def calc_Z_sum(composition):
    '''
    Calculates the sum of the atomic numbers of all atoms in a
    compositional unit

    Args:
        composition - dictionary of atomic elements with values as tuples of
        the form (Z,charge,n)
    Returns:
        z_tot - total atomic number
    '''
    # Loop and tuple unpacking used over indices and list comprehension for
    # readability
    # Equivalent to:
    # sum(comp[el][0]*comp[el][2] for el in comp)
    # where comp = composition and el = element
    Z_tot = 0
    for element in composition:
        # Unpack Z and n
        Z, _, n = composition[element]
        Z_tot += (Z * n)
    return Z_tot


@lru_cache(maxsize=1)
def load_ff_data():
    with importlib_resources.open_binary('LiquidDiffract.resources', 'form_factors.npy') as fp:
        ff_data = np.load(fp, allow_pickle=True)
    return ff_data


@data_cache(maxsize=32)
def calc_atomic_ff(element, Q):
    '''
    Calculates the atomic form factors fp(Q) of an element, p, over the
    Q values, Q. Uses the analytic tabulation published in the International
    Tables of Crystallography.

    This data is also freely available at:
    http://lampx.tugraz.at/~hadley/ss1/crystaldiffraction/atomicformfactors/formfactors.php

    Args:
        element - tuple of form (Z, charge, *_)
        Q - Data in Q space and units Angstroms^-1

    Returns:
        form_factor - NumPy array of atomic form factors at each Q value

    note: calc_atomic_ff ignores n values - atomic fractions or number of
          atoms in formula unit must be treated elsewhere as the function
          calculates form factor for 1 atom in the compositional unit.
    '''
    atomic_number, charge, *_ = element

    # Data is loaded in a separate function so it can be cached
    ff_data = np.copy(load_ff_data())
    ff_data = ff_data[np.where((ff_data[:, 0] == charge) &
                               (ff_data[:, 10] == atomic_number))][0, 1:10]
    # s = Q / 4pi, s here is s^2
    s = (Q / (4 * np.pi))**2
    form_factor = (ff_data[0] * np.exp(-ff_data[1] * s) +
                   ff_data[2] * np.exp(-ff_data[3] * s) +
                   ff_data[4] * np.exp(-ff_data[5] * s) +
                   ff_data[6] * np.exp(-ff_data[7] * s) + ff_data[8]
                   )
    return form_factor


def calc_effective_ff(composition, Q):
    '''
    Calculates the effective electronic form factor as defined in Eq. 10 of
    Eggert et al., 2002.

    i.e.

    f_e(Q) = SUM f_p(Q) / Z_tot
    where f_p(Q) are the atomic form factors and Z_tot is the total atomic
    number of the compositional unit

    Args:
        composition - composition dictionary where values are tuples of the
        form (Z, charge, n)
        Q - q space values to evaluate over

    Returns:
        effective_ff - effective electronic form factor at Q values
        atomic_ff - atomic form factors at Q values for each element
    '''
    Z_tot = calc_Z_sum(composition)
    # Expand composition dictionary to list of tuples with entries repeated
    # where n>1
    composition_atoms = [tuple(element) for expanded in
                         [[composition[species]]*composition[species][2]
                          for species in composition]
                         for element in expanded]
    atomic_ff = []
    for atom in composition_atoms:
        atomic_ff.append(np.copy(calc_atomic_ff(atom, Q)))
    atomic_ff = np.asarray(atomic_ff)
    effective_ff = np.sum(atomic_ff, 0) / Z_tot
    return effective_ff, atomic_ff


def calc_average_scattering(composition, Q):
    '''
    Computes the average scattering functions, <f>2 and <f2> required for the
    faber-ziman formalism of total structure factor S(Q). <f>2 is the square of
    the mean scattering factors, whereas <f2> is the mean square average.

    <f2> = 1/N Σ_p (f_p(Q))^2
    <f>2 = 1/N**2 Σ_p Σ_q f_p(Q) * f_q(Q)

    where:
        N is the number of atoms in the composition (e.g. MgSiO3: N = 5)
        f_p(Q) are the atomic atomic form factors of an element, p, calculated
        over the Q values, Q

    <f2> is equal to <f>2 for the monatomic case

    Args:
        composition - dictionary with keys that are element symbols and values
                      are tuples of the form (Z, charge, n_atoms)
        Q - Q-space values to evaluate over

    Returns:
        average_scattering - list of average scattering functions
            average_scattering[0] --> <f2>
            average_scattering[1] --> <f>2
    '''
    # Expand composition dictionary to list of tuples with entries repeated
    # where n>1
    composition_atoms = [tuple(element) for expanded in
                         [[composition[species]]*composition[species][2]
                          for species in composition]
                         for element in expanded]
    N = len(composition_atoms)
    atomic_ff = []
    for atom in composition_atoms:
        atomic_ff.append(np.copy(calc_atomic_ff(atom, Q)))
    atomic_ff = np.asarray(atomic_ff)

    # Create list to hold both function
    average_scattering = []

    # 1st function
    average_scattering.append(
            1/N * np.sum(atomic_ff**2, 0)
            )
    # 2nd function
    average_scattering.append(
            1/N**2 *
            np.sum(np.asarray([x*y for x in atomic_ff for y in atomic_ff]), 0)
            )
    return average_scattering


@lru_cache(maxsize=8)
def load_compton_data(element):
    element_data_fname = element + '.npy'
    with importlib_resources.open_binary('LiquidDiffract.resources.hubbel_compton', element_data_fname) as fp:
        cs_Q, _, cs_comp = np.load(fp, allow_pickle=True)
    return cs_Q, cs_comp


@data_cache(maxsize=32)
def calc_compton_scattering(element, Q):
    '''
    Interpolates compton scattering from tabulated data to estimate scattering
    intensity at each Q value.

    The analytical tabulations of compton scattering intensity are taken
    from Hubbel et al., 1975, J. Phys. Chem. Ref. Data 4, 471.

    Args:
        element - dictionary entry from 'comp' where keys are element symbols
        Q - Q values to estimate compton scattering intensity at
    '''
    # Element specific compton data is cached to save i/o calls
    cs_Q, cs_comp = load_compton_data(element)
    cs_Q, cs_comp = np.copy(cs_Q), np.copy(cs_comp)

    # The data is tabulated using s values instead of Q > Q = 4*pi*s
    cs_Q *= (np.pi * 4)
    # Interpolate compton scattering to required Q spacings
    finterp = scipy.interpolate.interp1d(cs_Q, cs_comp, kind='cubic', fill_value='extrapolate')
    return finterp(Q)


def calc_total_compton_scattering(composition, Q):
    '''
    Helper function to calculate total compton (incoherent) scattering for
    all atoms in a formula unit (composition)

    See function core.calc_compton_scattering for details on data used
    '''
    compton_scattering = []
    for element in composition:
        compton_scattering.append(np.copy(calc_compton_scattering(element, Q)) *
                                  composition[element][2])
    compton_scattering = np.sum(np.asarray(compton_scattering), 0)
    return compton_scattering


def calc_J(composition, Q):
    '''
    Calculates J value as defined in Eq 35 of Eggert et al., 2002
    i.e.
        J(Q) = SUM {Compton(Q)} / Z_tot^2 * f_e(Q)^2
        where f_e(Q) is the effective electronic form factor

    Args:
        composition - dictionary where keys are element symbols and values
        are tuples of the form (Z, charge, n)

        Q - Q values to evaluate
    '''
    effective_ff, _ = calc_effective_ff(composition, Q)
    Z_tot = calc_Z_sum(composition)
    compton_scattering = calc_total_compton_scattering(composition, Q)
    J = compton_scattering / (Z_tot**2 * effective_ff**2)
    return J


def calc_K_p(composition, Q):
    '''
    Calculates average effective atomic numbers of each atomic element
    in a formula unit over a set Q range as defined in Eqs 11 & 14 of
    Eggert et al., 2002.

    i.e.

    K_p(Q) for each element = f_p(Q)/f_e(Q)
    K_p for each element = <K_p(Q)>_Q

    Where:
        f_e(Q) is the effective electronic form factor of the total composition
        f_p(Q) are the atomic form factors for each atom, p, in the formula unit
        K_p(Q) is the Q-dependent effective atomic number for each atom, p
        K_p is the average effective atomic number


    Args:
        composition - dictionary of elements in composition where values
        are tuples of the form (Z, charge, n)

        Q - Q range to evaluate over

    Returns:
        K_p - numpy array of K_p values with length equal to number of atoms
              in a formula unit of the composition
    '''
    effective_ff, atomic_ff = calc_effective_ff(composition, Q)
    K_p = atomic_ff / effective_ff
    K_p = np.mean(K_p, 1)
    return K_p


def calculate_weights(composition, Q):
    '''
    Calculate X-ray scattering weightings for atomic pair correlations in
    the composition for a set Q-range. Uses the Warren-Krutter-Morningstar
    approximation for average effective atomic number.
    '''
    # Get average effective atomic numbers for each atom
    K_p_list = calc_K_p(composition, Q)
    # Number of atoms in composition
    N = len(K_p_list)
    # K_p_norm - normalise to number of atoms in composition
    K_p_norm = (np.sum(K_p_list)/N)**2
    # List of atoms in composition
    atom_list = [atom for expanded in
                 [[species]*composition[species][2]
                  for species in composition.keys()]
                 for atom in expanded]
    # Dict of Kp for each atom species
    K_p_dict = dict(zip(atom_list, K_p_list))
    # List of atom (alpha-beta) pairs
    pairs = list(set(cartesian_product(atom_list, repeat=2)))
    weights = {}
    c_dict = {}

    for pair in pairs:
        alpha = pair[0]
        beta = pair[1]
        c_alpha = composition[alpha][2]/N
        c_beta = composition[beta][2]/N
        K_alpha = K_p_dict[alpha]
        K_beta = K_p_dict[beta]
        # Kronecker delta - i.e. delta = 1 of a==b and delta = 0 if a!=b
        delta = int(alpha == beta)
        # Calculate weight for alpha-beta pair
        w_alphabeta = (c_alpha*K_alpha * c_beta*K_beta * (2-delta))/K_p_norm

        weights[(alpha,beta)] = w_alphabeta

    for atom in composition:
        c_beta = composition[atom][2]/N
        c_dict[atom] = c_beta

    return weights, c_dict


def calc_S_inf(composition, Q, method='ashcroft-langreth'):
    '''
    Caclulates the value S_inf as defined in Eq 19 of Eggert et al., 2002
    i.e. S_inf = SUM K^2_p / Z_tot^2
    S_inf defaults to 1 if the composition is monatomic

    For the Faber-Ziman structure factor formulation, the interference
    function, i(Q) is defined as S(Q) - 1. calc_S_inf is still called in the
    program to aid readability and avoid too many if statement blocks in the
    main code, though there is a slight performance decrease.
    '''
    if method == 'faber-ziman':
        S_inf = 1.0

    elif method == 'ashcroft-langreth':
        N = np.sum([composition[el][2] for el in composition])
        if N == 1:
            S_inf = 1.0
        elif N > 1:
            K_p = calc_K_p(composition, Q)
            Z_tot = calc_Z_sum(composition)
            S_inf = np.sum(K_p**2) / Z_tot**2
        else:
            raise ValueError('Error in composition - n value < 1')

    else:
        raise ValueError('Please select a valid method for structure factor')
    return S_inf


def calc_alpha(Q_cor, I_cor, rho,
               Z_tot=None, J=None, S_inf=None, effective_ff=None,
               average_scattering=None, compton_scattering=None,
               method='ashcroft-langreth'):
    '''
    Calculates 'alpha' - the normalisation factor

    For the Ashcroft-Langreth formalisation this is defined in
    Eq.34 of Eggert et al., 2002.

    For the Faber-Ziman formalism this is defined in equation A13 of
    Morard, 2013 but note there is a missing bracket - refer to the definition
    in Decremps et al., 2016, PhysRev B 93, 054209 instead.

    Args:
        Q_cor - Rebinned Q space data cut at Q_max
        I_cor - Background corrected intensity data
        rho - actual or initial estimate of atomic density in atoms/Angstrom^3
        method - 'ashcroft-lamgreth' or 'faber-ziman'

        For AL:
        ------
        Z_tot - Atomic number sum of elements in composition
        J - J value as defined in Eq 35 of Eggert et al., 2002
        S_inf - S_inf value as defined in Eq 19 of Eggert et al., 2002
        effective_ff - effective electronic form factor

        For FZ:
        ------
        average_scattering - Average scattering functions <f2> and <f>2
                             see calc_average_scattering for definitions
        compton_scattering - Compton scattering intensity

    Returns:
        alpha
    '''
    if method == 'ashcroft-langreth':
        # define the two integrals in the equation
        int_1 = simps((J + S_inf) * Q_cor**2, Q_cor)
        int_2 = simps((I_cor / effective_ff**2) * Q_cor**2, Q_cor)
        # calculate alpha
        alpha = Z_tot**2 * (((-2 * np.pi**2 * rho) + int_1) / int_2)
        return alpha
    elif method == 'faber-ziman':
        # define the two integrals in the equation
        int_1 = simps(((compton_scattering + average_scattering[0])/average_scattering[1]) * Q_cor**2, Q_cor)
        int_2 = simps(((Q_cor**2 * I_cor) / average_scattering[1]), Q_cor)
        # Calculate alpha
        alpha = ((-2 * np.pi**2 * rho) + int_1)/int_2
        return alpha
    else:
        raise ValueError()


def calc_coherent_scattering(Q_sample, I_sample, composition, alpha,
                             compton_scattering=None, method='ashcroft-langreth'):
    '''
    Calculates coherent scattering intensity for a given composition

    Coherent scattering is defined as:
    Coh_I = N * [alpha * Sample_scattering(Q) - SUM Incoherent_scattering]
    where N is number of atoms, n_atoms

    For the Faber-Ziman formalism there is no need to scale by the number
    of atoms in the compositional unit (N)

    Args:
        Q_sample - The Q-values to evaluate over (np array)
        I_sample - The sample scattering intensity (i.e. bkgd corrected) (np array)
        composition - Dict of sample composition
        alpha - The normalisation factor
        method - S(Q) formalism, 'ashcroft-langreth' or 'faber-ziman'

    Optional:
        compton_scattering - The compton scattering intensity for the sample
    '''
    # If compton scattering data is not provided, retrieve it now
    if compton_scattering is None:
        compton_scattering = calc_total_compton_scattering(composition, Q_sample)
        if method == 'faber-ziman':
             compton_scattering /= sum(composition[el][2] for el in composition)
    else:
        pass
    coherent_scattering = (alpha * I_sample) - compton_scattering
    if method=='ashcroft-langreth':
        # N - number of atoms in compositional unit (molecule)
        n_atoms = sum(composition[el][2] for el in composition)
        # Normalise coherent scattering
        coherent_scattering *= n_atoms
    else:
        pass

    return coherent_scattering


def calc_structure_factor(Q_cor, I_cor, composition, rho, method='ashcroft-langreth'):
    '''
    Calculates the molecular structure factor in normalised units using the
    method first developed by Krogh-Moe (1956) and Norman (1957).

    Args:
        Q_cor - Rebinned Q space data
        I_cor - Bkgd corrected intensity data
        composition - dictionary of atomic elements, with values as tuples in
                      the form (Z,charge,n)
        rho - actual or initial estimate of atomic density in atoms/Angstrom^3

    The code provides an option to use either the Ashcroft-Langreth or
    Faber-Ziman formulation of the structure factor S(Q)

    S_AL(Q) = I^COHERENT(Q) / N * Z_TOT^2 * f_e^2(Q)

    S_FZ(Q) = I^COHERENT(Q) - (<f^2> - <f>^2 ) / <f>^2

    I^COHERENT (the coherent scattering component of measured scattering
    intensity) has a different formulation for AL & FZ methods.

    I^COH_AL = N [alpha_AL * I^SAMPLE(Q) - SUM_p[I^COMPTON_p(Q)] ]

    I^COH_FZ = alpha_FZ * I^SAMPLE(Q) - SUM_p[I^COMPTON_p(Q)]

    I^COMPTON_p is the compton (incoherent) scattering of each atom
    (see core.calc_compton_scattering)

    alpha is the normalisation factor (see core.calc_alpha)

    N is the total number of atomic species in a formula unit
    (e.g. for SiO2 N=3)

    Z_TOT is the total Z number of a formula unit

    f_e(Q) is the effective electronic form factor

    <f^2> and <f>^2 are intermediate functions which quantify average
    scattering in a similar way to the effective electronic form factor used
    in the Ashcroft-Langreth formulation
    '''
    # N is number of atoms in 1 formula unit
    n_atoms = sum(composition[element][2] for element in composition)
    if method == 'ashcroft-langreth':
        # Z is total atomic number of formula unit
        Z_tot = calc_Z_sum(composition)
        # Calculate intermediate functions for alpha/coherent scattering for
        # computational efficiency
        J = calc_J(composition, Q_cor)
        S_inf = calc_S_inf(composition, Q_cor, method=method)
        # effective electronic form factor also used to calculate alpha
        effective_ff, _ = calc_effective_ff(composition, Q_cor)
        alpha = calc_alpha(Q_cor, I_cor, rho,
                           Z_tot=Z_tot, J=J, S_inf=S_inf,
                           effective_ff=effective_ff, method='ashcroft-langreth')
        coherent_scattering = calc_coherent_scattering(Q_cor, I_cor, composition, alpha, method=method)
        structure_factor = coherent_scattering / (n_atoms * Z_tot**2 * effective_ff**2)
        return structure_factor

    elif method == 'faber-ziman':
        # Calculate average scattering functions <f^2> & <f>^2
        avg_scattering_f = calc_average_scattering(composition, Q_cor)
        # Calculate compton (incoherent scattering) used in alpha_FZ
        compton_scattering = calc_total_compton_scattering(composition, Q_cor)
        compton_scattering = compton_scattering / n_atoms
        # Calculate alpha
        alpha = calc_alpha(Q_cor, I_cor, rho,
                           average_scattering=avg_scattering_f,
                           compton_scattering=compton_scattering,
                           method='faber-ziman')
        # Calculate coherent scattering
        coherent_scattering = calc_coherent_scattering(Q_cor, I_cor,
                                                       composition, alpha,
                                                       compton_scattering=compton_scattering,
                                                       method=method)
        # Calculate structure factor
        structure_factor = (coherent_scattering -
                            (avg_scattering_f[0] - avg_scattering_f[1])
                           ) / avg_scattering_f[1]
        return structure_factor

    else:
        raise ValueError('Please select a valid method for structure factor')

def get_mod_func(q, mod_func, window_start):
    '''
    Returns the modification function to scale S(q) before FFT

    Implemented modification functions:
    ----------------------------------

    'lorch' - from Lorch, 1969:

        M(Q) = sin(pi*Q/Q_max) / (pi*Q/Q_max)

    'cosine-window' -  - from Drewitt et al., 2013:

        M(Q) = 1 if Q < Q1
        M(Q) = 0.5[1 + cos(x*pi / N-1)] if Q1 <= Q <= Qmax

        where N is width of the window function
        and x is an integer with values from 0 - (N-1) across the window

        if cosine-window is selected, optional kwarg window-start must
        be passed (Q1)
    '''
    if mod_func == 'Cosine-window':
        pre_window = np.ones(len(q[np.where(q<window_start)]))
        window = np.arange(len(q[np.where(q>=window_start)]))
        window_length = len(window)
        modification = 0.5 * (1 + np.cos(window*np.pi/(window_length-1)))
        modification = np.concatenate((pre_window, modification))

    elif mod_func == 'Lorch':
        with np.errstate(divide='ignore', invalid='ignore'):
            modification = np.sin(np.pi * q / np.max(q)) / (np.pi * q / np.max(q))
    else:
        modification = 1

    return modification


def calc_correlation_func(x, y, rho, dx='check', N=12, mod_func=None,
             window_start=None, function='density_func'):
    '''
    Calculates the fourier transform of the interference function
    i(Q) = S(Q) - S_inf to obtain the real-space correlations functions:
    1) pair-distribution function [g(r)], 2) the differential correlation
    function (density function) [D(r)] or, 3) the radial distribution function
    [RDF(r)]. The conversion to real-space is useful for structural (PDF) analysis.

    Example:

    F(q) --> G(r) transformation
    FFT here is sine-only and so assumes F(q) is an odd function
    F(q) is therefore padded with its odd image and zero-padded in the middle
    for efficiency and for interpolation in r-space (padded len =  2^n).

    i.e. F(q) = [0 1 3 2] --> F(q)_pad = [0 1 3 2 0 -2 -3 -1]

    Values of G(r) are the imaginary components of the fourier transorm of
    F(q). The real components are all zero. Only the first half of the
    transformed array is used as the second is its odd image.

    dr correponds to the minimum frequency available for the entire Q-range.
    i.e.:
        dr = 2 * np.pi / (len(F(q)_padded) * dq)

    Therefore for the reverse transformation
        dq = np.pi / (len(r) * dr)

    The pair-distribution function is proportional to the probability of
    finding an atom at a distance r from an average atom taken as the origin.
    The radial distribution function gives atomic coordination numbers when
    integratd across peaks. The density function is used in the analysis here
    to normalise i(q).

    Pair-distribution function g(r):

        g(r) - 1 = 1/(2 * pi^2 * r * rho_0) * INT[ q[i(q)] sin(qr) dq]

    Differential correlation function (density function) D(r) = F(r):

        D(r) = 4 * pi * r * rho_0 * [g(r) - 1]
        D(r) = 2/pi * INT[ q[i(q)] sin(qr) dq]

    Radial distribution function, RDF(r):

        RDF(r) = 4 * pi * r^2 * rho_0 * g(r)
        RDF(r) = (2r/pi)*INT[ q[i(q)] sin(qr) dq] + 4*pi*r^2*rho_0

    Args:
        x   - Q values (numpy array)
        y   - Corresponding values of the interference function (numpy array)
        rho - atomic density (float)
        dq  - steps of q values (which need to be constant)
              this defaults to 0.02, but can also be calculated in function if
              set to 'check'
        N   - Sets size of array for fft where array_size = 2^N, defaults to 12
        mod_func - Modification function to scale S(Q) before FFT
        function - density_func, pair_dist_func, radial_dist_func

    Returns:
        r   - Array of r values
        D_r/g_r/RDF_r - Corresponding array of values for calculated function
    '''
    # Define dx (qstep) if not passed to function
    if dx == 'check':
        dx = x[1] - x[0]
    else:
        pass

    # Get modification function
    modification = get_mod_func(x, mod_func, window_start)

    # Calculate qi(q) input array
    z = x * y * modification
    del modification
    # Make sure using float64
    z = z.astype(np.float64, copy=False)
    # Pad with enough zeros to make array of size 2**N
    # Length of array to fourier transform is 2**N
    # The array is mirrored and padding zeros are added to increase
    # the size of the array
    # Therefore len(z) < ((2**N)/2)
    # N=12 allows max 2048 data points (at q-spacing 0f 0.02 range is 40.94)
    if len(z) > (2**N/2):
        # Error checking also takes place in LiquidDiffract.gui
        # so that the graphical app does not crash
        raise ValueError('Length of data array > [(2**N)/2]\n' +
                         '            Please check your data is in ' +
                         'Q-space or increase value of \'N\'')
    else:
        pass
    padding_zeros = np.int(2**N - (len(z)*2 - 1))
    # Pad array with odd image of function
    z = np.concatenate((z, np.zeros(padding_zeros), -np.flip(z[1::], 0)))
    # Fourier transform here uses scipy.fftpack over numpy.fft
    # ifft used as converting from frequency (intensity) domain
    # This means no array-length scaling is needed and the result is not negative
    fft_z = scipy.fftpack.ifft(z)
    # Only the imaginary component of the transformed array is used
    # Only the first half of the array is used as the rest is the odd component
    fft_z = np.imag(fft_z)[:len(fft_z)//2]
    # Scaling for the inverse transform is QMax = dx * (2**N)/2
    fft_scaling = 2**N / 2 * dx
    fft_z *= fft_scaling
    # Calculate steps of r and r array
    dr = np.pi*2/(len(z)*dx)
    r = np.arange(len(fft_z))*dr
    if function == 'density_func':
        # Scale by factor 2/pi
        D_r = fft_z * 2/np.pi
        return r, D_r
    elif function == 'pair_dist_func':
        with np.errstate(divide='ignore', invalid='ignore'):
            sf = 1 / (2 * r * rho * np.pi**2)
            g_r = fft_z * sf
        g_r += 1
        return r, g_r
    elif function == 'radial_dist_func':
        RDF_r = (2 * r * fft_z / np.pi) + (4 * np.pi * rho * r**2)
        return r, RDF_r
    else:
        raise ValueError('arg \'function\' must be valid option')

def normalise_achroft_langreth_func(S_inf, sq=None, gr=None, rdf=None,
                                    rho_0=None, r=None):
    '''
    Helper function to normalise functions in the Aschroft-Langreth formalism
    for polyatomic samples. For polyatomic samples 0 < S_inf < 1 and so
    S(Q) --> S_inf at infinite Q and g(r) --> (1-S_inf) at r=0. RDF(r) and T(r)
    are similarly scaled. Renormalising these functions by S_inf makes
    comparisons to functions in the Faber-Ziman formalism easier and allows the
    same curve-fitting routines to be used on both.

    S(Q) normalisation:

        S_s(Q) = S_AL(Q) / S_inf

    g(r) normalisation:

        g_s(r) = (g(r) - 1)/S_inf + 1

    RDF(r) normalisation:

        RDF_s(r) = (RDF(r) - 4πr^2ρ_0)/S_inf + 4πr^2ρ_0

    Args:
        S_inf   - S_inf value for normalisation (see core.calc_S_inf)
        sq*     - Values of S(Q) to normalise (numpy array)
        gr*     - Values of g(r) to normalise (numpy array)
        rdf**   - Values of RDF(r) to normalise
        rho_0*  - atomic number density (float)
        r*      - Values of r that correspond to the RDF(r)

        *  - Optional kwargs
        ** - The RDF(r) normalisation requires the additional kwargs rho_0 & r

    Returns 1 of:

        sq_norm  - Renormalised S(Q) (numpy array)
        gr_norm  - Renormalised g(r) (numpy array)
        rdf_norm - Renormalised RDF(r) (numpy array)
    '''
    # S(Q) normalisation
    if sq is not None:
        sq_norm = sq / S_inf
        return sq_norm
    elif gr is not None:
        gr_norm = (gr - 1)/S_inf + 1
        return gr_norm
    elif rdf is not None:
        try:
            constant = 4*np.pi*rho_0*r**2
            rdf_norm = (rdf - constant)/S_inf + constant
            return rdf_norm
        except TypeError:
            raise NameError('RDF(r) normalisation requires keyword arguments rho_0 and r')
    else:
        raise NameError('Must provide function to normalise')


def calc_D_r_iteration_term(delta_D_r, N=12, dq=0.02):
    '''
    Calculates the term responsible for oscillations at small r values.
    Equation 45 in Eggert et al., 2002

    i.e. Delta_alpha * Q * S_inf = INT 0>r_min [Delta_D(r) sin(Qr) dr]

    Delta_D(r) is the difference between D(r) and its expected behaviour

    N defines the size of the array to be fourier transormed (2**N)
    This should be the same as used to calculate D(r) from i(Q) to
    preserve correct scaling factors
    '''
    # Make sure array size is not too large for padding value N
    if len(delta_D_r) > (2**N/2):
        raise ValueError('Length of array > [(2**N)/2]')
    else:
        pass
    padding_zeros = np.int(2**N - (len(delta_D_r)*2 - 1))
    z = np.concatenate((delta_D_r,
                        np.zeros(padding_zeros),
                        -np.flip(delta_D_r[1::], 0)))
    # For the forward fourier transform we need the negative of the imaginary components
    # in the first half of the array
    fft_z = -np.imag(scipy.fftpack.fft(z))
    # The fourier transform of Qi(Q) was scaled by the artificial Q_max to obtain the corrected
    # magnitudes - i.e. fft_scaling = len(D(r) * dQ = 2**N / 2 * q_step
    # This scaling has to be reversed here. For N=12 and q_step = 0.02 the scaling is 40.96
    fft_scaling = 2**N / 2 * dq
    fft_z = fft_z / fft_scaling
    # For the forward transform here
    return fft_z


@data_cache(maxsize=128)
def calc_model_D_intra_r(rho, r):
    '''
    Returns the model behaviour of D(r) at r < r_min, where r_min is the
    largest distance (0 -- r-min) where no atom can be found.
    In a liquid, this should be the distance of the 1st coordination shell.

    The model behaviour is:

         D_(r<r_min) = D_intra(r) − 4πrρ

         D_intra(r) is the expected contribution of *intra*molecular forces
         i.e. interaction of atoms within the same molecule

         This can be calculated using a fixed model approach
         (see eq. 42 in Eggert et al., 2002)

         There are also further intermolecular contributions at r<r_min
         if the sample has intermolecular coordination between compositional
         units (e.g. polymeric liquids or framework glasses)

     LiquidDiffract assumes D_intra(r) is negligible. This is only true for
     some compositions (e.g. monatomic metallic liquids). This functionality
     may be added in future releases however.
    '''
    model_D_intra_r = -4 * np.pi * r * rho
    return model_D_intra_r


def calc_chi_squared(r_intra, delta_D_r, method='simps'):
    '''
    Calculates a chi squared figure of merit as in equation 50 of Eggert. This
    is a measure of how small Delta_D_i(r) is, to be used as a tolerance factor
    to check convergence of i_i+1(q) and in the optimisation routine of rho/s.

    Chi^2 = INT 0-r_min [Delta_D(r)]^2 dr

    The value chi^2 is defined as the area under the curve Delta_D(r) limited
    to r_min (squared to remove negatives).

    The Chi^2 value used in LiquidDiffract is scaled by 1e6 to enable easier minimisation.

    Args:
        delta_D_r - numpy array of delta_D_r from r=0 to r=r_min

    Returns:
        chi_squared - Value of chi^2
    '''
    # Square to remove negatives
    f = delta_D_r ** 2
    if method == 'simps':
        # Integrate using simpson's rule
        chi_squared = simps(f, x=r_intra)
    else:
        raise NotImplementedError()
    return chi_squared * 1e6


def stop_iteration(stop_condition='count', count=None,
                   iter_limit=5, chi_squared=None):
    '''Helper function for checking stop condition of iteration

    Currently LiquidDiffract uses an iteration limit only so this function
    is very unnecessary/a little expensive. It is left here to make any changes
    easier to implement (i.e. checking for chi^2 convergence)
    '''
    if stop_condition == 'count':
        if count < iter_limit:
            return False
        else:
            return True
    else:
        raise NotImplementedError()


def calc_impr_interference_func(q_data, interference_func,
                                composition, rho,
                                r_min, iter_limit,
                                method, mod_func, window_start, fft_N):
    '''
    Calculates an improved estimate of the interference function via the
    iterative procedure described in Eggert et al., 2002. This is done by
    scaling the data and forcing its behaviour in the low r region to some
    modelled behaviour.

    The function also returns a chi^2 figure of merit which can be passed
    to a minimisation routine to estimate rho/b.

    Args:
        q_data - Numpy array of Q space data
        interference_func - Initial interference function
        composition - Dict of elements, values as tuples of form (Z,charge,n)
        rho - Density (average number density in atoms/angstrom^3)
        r_min - Intramolecular distance cut-off
        iter_limit - Iteration limit for S(Q) refinement. iter_limit >= 3 is
                     recommended for convergence. If optimising for minimum
                     chi^2 then 3 <= iter_limit <= 10 is recommended. Please
                     see the LiquidDiffract README for more information
        method - Formalisation of total scattering structure factor to use
                 ('ashcroft-langreth' or 'faber-ziman')
        mod_func - Modification function to scale data before fft
                   None, 'Cosine-window' or 'Lorch'
        window_start - Start of Cosine-window function needed if
                       mod_func == 'Cosine-window'
        fft_N - Sets size of array for fft where array_size = 2^N

    Returns:
        interference_func_impr - Interference function after iter_limit iterations
        chi_sq - Figure of merit (see calc_chi_squared)
    '''
    # Unpack rho - solver passes rho as a single element array
    rho = np.float64(rho)

    dq = q_data[1] - q_data[0]
    # Calculate initial D(r)
    r, D_r = calc_correlation_func(q_data, interference_func, rho,
                      mod_func=mod_func, window_start=window_start,
                      dx=dq, N=fft_N)
    # Calculate expected behaviour of D(r) for intramolecular distances (r<r_min)
    # and scale by S_inf (1 for FZ/monatomic - only necessary for AL-S(Q))
    model_D_intra_r = (np.copy(calc_model_D_intra_r(rho, r)) *
                       calc_S_inf(composition, q_data, method=method))

    # Calculate static terms of iterative proceduce
    with np.errstate(divide='ignore', invalid='ignore'):
        t1 = 1 / q_data

    if method == 'ashcroft-langreth':
        t2_divisor = calc_S_inf(composition, q_data, method=method) + calc_J(composition, q_data)
    elif method == 'faber-ziman':
        t2_divisor = 1.0

    # Count no. iterations for verbosity or control
    # Function set up to call stop_looping func in case want to modify
    # to use chi-squared value as tolerance value instead
    count = 1
    done_looping = False
    while not done_looping:
        try:
            done_looping = stop_iteration(stop_condition='count', count=count, iter_limit=iter_limit)
            interference_func = interference_func_impr
            r, D_r = calc_correlation_func(q_data, interference_func, rho,
                              mod_func=mod_func, window_start=window_start,
                              dx=dq, N=fft_N)
        except NameError:
            pass

        # Calculate the difference between observed and model D(r) along
        # with the chi^2 figure of merit
        delta_D_r = (D_r - model_D_intra_r)[np.where(r < r_min)]
        r_intra = r[np.where(r<r_min)]
        chi_squared = calc_chi_squared(r_intra, delta_D_r)

        # Calculate next iteration of i(Q)
        t3 = calc_D_r_iteration_term(delta_D_r, N=fft_N, dq=dq)[:len(interference_func)]
        t2 = (interference_func/t2_divisor) + 1
        with np.errstate(invalid='ignore'):
            interference_func_impr = interference_func - (t1 * t2 * t3)

        count += 1

    return interference_func_impr, chi_squared


def refinement_objfun(minvar, *args):
    '''
    Wrapper objective function to pass to a solver to refine density (rho)
    and/or background scaling factor (b) by minimising a chi^2 figure of merit.

    This function provides set-up and logic for dealing with refinement of
    different variables, including recalculating the background subtraction
    when refining b. The actual calculations for the optimisation objective
    function are in core.calc_impr_interference_func

    Args:

        minvar - independent variable(s) in the optimisation (numpy array with shape (n,))
                 Can be one or both of:
                    rho - average (bulk) number density (atoms/angstrom^3)
                    bkg_scale - background scaling factor

        *args - The expected arguments depend on if the background scaling factor
                is being refined. The last two arguments should be opt_rho and
                opt_bkg, i.e.:

                    *_, opt_rho, opt_bkg = args

                These should be boolean flags indicating if the density and
                background scale factor are being refined respectively.

                If opt_bkg == True (1), the expected args are:

                    (*_, q_data, I_data_uncorrected, I_bkg,
                     qmin, qmax,
                     smooth_flag, smooth_window_length, smooth_poly_order,
                     composition, r_min, iter_limit,
                     method, mod_func, window_start, fft_N,
                     opt_rho, opt_bkg) = args

                If opt_rho == False (0), the density should precede q_data

                If opt_bkg == False (0), the expected args are:

                    (q_data, I_data_corrected,
                     composition, r_min, iter_limit,
                     method, mod_func, window_start, fft_N,
                     opt_rho, opt_bkg) = args

            The individual arguments are:

                q_data - Numpy array of Q space data

                I_data_uncorrected - Numpy array of intensity (I(Q)) values
                                     prior to background subtraction
                I_data_corrected - Numpy array of background subtracted I(Q) values
                I_bkg - Numpy array of background intensities
                qmin - Minimum Q value of useable data (float or None)
                qmax - Maximum Q value to truncate data (float or None)
                smooth_flag - Boolean flag to apply savitzky golay filter
                smooth_window_length - Window length for optional smoothing filter
                smooth_poly_order - Polynomial order for optional smoothing filter
                composition - Dict of elements, values as tuples of form (Z,charge,n)
                r_min - Intramolecular distance cut-off
                iter_limit - Iteration limit for S(Q) refinement. iter_limit >= 3 is
                             recommended for convergence. If optimising for minimum
                             chi^2 then 3 <= iter_limit <= 10 is recommended. Please
                             see the LiquidDiffract README for more information
                method - Formalisation of total scattering structure factor to use
                         ('ashcroft-langreth' or 'faber-ziman')
                mod_func - Modification function to scale data before fft
                           None, 'Cosine-window' or 'Lorch'
                window_start - Start of Cosine-window function needed if
                               mod_func == 'Cosine-window'
                fft_N - Sets size of array for fft where array_size = 2^N

    Returns:
        chi_sq - Figure of merit (see calc_chi_squared)
                 This is the value to be minimised by optimising
                 the independent variable(s) minvar
    '''
    *_, opt_rho, opt_bkg = args
    # Unpack optimisation variables (rho and/or bkg_scale)
    if opt_rho == True and opt_bkg == True:
        rho = np.float64(minvar[0])
        bkg_scale = np.float64(minvar[1])
    # Unpack rho - solver passes rho as a single element array
    elif opt_rho == True and opt_bkg == False:
        rho = np.float64(minvar)
    # Unpack bkg_scale as abovs
    elif opt_rho == False and opt_bkg == True:
        bkg_scale = np.float64(minvar)
    else:
        raise NotImplementedError()

    if opt_rho == False:
        rho, *_ = args
        rho = np.float64(rho)

    if opt_bkg == True:
        # Unpack args - additional args needed when refining bkg_scale
        (*_,
         q_data, I_data_uncorrected, I_bkg,
         qmin, qmax, smooth_flag, smooth_window_length, smooth_poly_order,
         composition, r_min, iter_limit,
         method, mod_func, window_start, fft_N, _, _) = args
        I_data_corrected = I_data_uncorrected - (I_bkg * bkg_scale)
        # Cut at qmax if qmax is != None
        if qmax:
            # Cut data at qmax
            cut_max = np.where(q_data < qmax)
            q_data = q_data[cut_max]
            I_data_corrected = I_data_corrected[cut_max]
        # Similarly cut at qmin if present
        if qmin:
            # Cut data to qmin - setting I(q<qmin)=I(qmin)
            fill_val = I_data_corrected[np.argmax(q_data > qmin)]
            cut_min = I_data_corrected[np.where(q_data > qmin)]
            padding = np.asarray([fill_val] * (len(q_data) - len(cut_min)))
            I_data_corrected = np.concatenate((padding, cut_min))
        # smooth data if smooth_flag == 1
        if smooth_flag == True:
            I_data_corrected = smooth_data(I_data_corrected,
                                      window_length=smooth_window_length,
                                      poly_order=smooth_poly_order)
    else:
        # Unpack args here for opt_bkg == False
        # (if opt_bkg is 0, opt_rho is always 1)
        (q_data, I_data_corrected, composition, r_min, iter_limit,
         method, mod_func, window_start, fft_N, _, _) = args

    structure_factor = calc_structure_factor(q_data, I_data_corrected,
                                             composition, rho,
                                             method=method)
    interference_func = structure_factor - calc_S_inf(composition, q_data, method=method)

    _, chi_squared = calc_impr_interference_func(q_data, interference_func,
                                                 composition, rho, r_min, iter_limit,
                                                 method, mod_func, window_start, fft_N)

    return chi_squared


def integrate_coordination_sphere(r, rdf,
                                  r_0=None, rp_max=None,
                                  r_max=None, r_min=None, method=0):
    '''
    Calculates the 1st coordination number, N_1, for a monatomic sample by 
    integrating the area under the first peak in RDF(r)

    The position of the 1st peak in RDF(r) directly gives the average 
    interatomic bond length, r_ab, between an atom 'a' at the origin and the
    next-nearest neighbour. The area under this curve gives the average 
    coordination number of the 1st coordination sphere, N_1. RDF(r) is the 
    radial distribution function, defined as:

        RDF(r) = 4πρr^2g(r), where g(r) is the pair-distribution function

    The coordination number is fairly sensitive to the upper integration limit 
    chosen. There are several methods used in the literature to define the 
    limits based on either physical models or empirical observations. Three 
    methods are implimented in LiquidDiffract:

        1) Integration of area symmetrical in rg(r)

        This method is based on the assumption that the quantity rg(r) is 
        symmetrical for a coordination shell about its average position. The
        coordination number is calculated by integrating across RDF(r) from 
        the leading edge of the 1st peak, r_0, to the peak centre of the
        function r*g(r), r'_max. The integral is doubled to account for the 
        right half of the peak:

            (Eq. 2.5.1 in Waseda, 1980)
            N_a = 2 * INT 4πρr[rg(r)]_sym dr | r=r_0 --> r=r'_max 

        2) Integration of area symmetrical in r^2g(r)

        This method is similar to method 1 but assumes the peak is symmetrical
        in r^2g(r) instead of rg(r). The limit r_max is therefore the peak 
        centre in RDF(r):

            (Eq. 2.5.2 in Waseda, 1980)
            N_b = 2 * INT 4πρ[r^2g(r)]_sym dr | r=r_0 --> r=r_max

        It should be noted that r_max > r'_max

        3) Integration to the firt minimum in RDF(r)

        This is the most commonly used method to estimate N_1. It is based on
        the observation that the 1st peak is generally assymetric in 
        experimental data. The function 4πρr^2g(r) is integrated from r_0 to 
        the first minimum after the peak:

            (Eq. 2.5.3 in Waseda, 1980)
            N_c = INT RDF(r) dr | r=r_0 --> r=r_min 

        N_c is sensitive to the value of r_min used and r_min is not always
        easily refined. Noise and truncation ripples in RDF(r) can make it
        difficult to determine r_min, as can the broadening of RDF(r) 
        associated with an increase in temperature.

    It should be noted that: N_a < N_b but N_c is not always > N_b
    N_a can be considered a lower-bound on the coordination numeber as the 
    peak is never truly symmetrical. In some cases N_c can provide an upper
    limit, but the value is much more sensitive as r_min is never as well
    defined as r_0 and r_max/r'_max

    Args:
        r   - r values (numpy array)
        rdf - Corresponding values of the radial distribution function, RDF(r) (numpy array)
        r_0 - Leading edge of the 1st peak in RDF(r)
        rp_max - r'_max, peak centre in rg(r)
        r_max - peak centre in r^2g(r)
        r_min - 1st minimum to the right of the 1st peak in RDF(r)

    Returns:
        N_a - 1st coordination number computed via method 1.
        N_b - 1st coordination number computed via method 2.
        N_c - 1st coordination number computed via method 3.
    '''

    rdf_interp = scipy.interpolate.interp1d(r, rdf, kind='cubic', fill_value='extrapolate')

    if method == 0:
        N_a = integrate_coordination_sphere(r, rdf, r_0=r_0, rp_max=rp_max, method=1)
        N_b = integrate_coordination_sphere(r, rdf, r_0=r_0, r_max=r_max, method=2)
        N_c = integrate_coordination_sphere(r, rdf, r_0=r_0, r_min=r_min, method=3)
        return N_a, N_b, N_c

    elif method == 1:
        N_a, _ = quadrature(rdf_interp, r_0, rp_max, maxiter=200)
        N_a *= 2.0
        return N_a

    elif method == 2:
        N_b, _ = quadrature(rdf_interp, r_0, r_max, maxiter=200)
        N_b *= 2.0
        return N_b

    elif method == 3:
        N_c, _ = quadrature(rdf_interp, r_0, r_min, maxiter=200)
        return N_c

    else:
        raise NameError('\'method\' must be one of either 1, 2, 3, or 0')



# -*- coding: utf-8 -*-
"""
GUI frontend for LiquidDiffract
<https://github.com/bjheinen/LiquidDiffract>
"""
__author__ = "Benedict J. Heinen"
__copyright__ = "Copyright 2018-2021, Benedict J. Heinen"
__email__ = "benedict.heinen@gmail.com"

import os.path
import webbrowser
# importlib.resources only available in python>=3.7
try:
    from importlib import resources as importlib_resources
except ImportError:
    import importlib_resources
import numpy as np
from PyQt5.QtWidgets import QMainWindow, QWidget, QVBoxLayout, QTabWidget, QAction
from PyQt5.QtGui import QIcon

from LiquidDiffract.gui import bkg_ui
from LiquidDiffract.gui import optim_ui
from LiquidDiffract.gui import results_ui
from LiquidDiffract.gui import structure_ui
from LiquidDiffract.gui import utility
from LiquidDiffract.version import __appname__, __version__


class App(QMainWindow):

    def __init__(self, screen_size):
        super().__init__()
        # Set options here before running initUI to initialise
        # Get primary (current) screen dimensions
        self.screen_size = screen_size
        self.width = self.screen_size.width()
        self.height = self.screen_size.height()
        self.initUI()

    def initUI(self):
        self.title = __appname__ + ' v' + __version__
        self.setWindowTitle(self.title)
        self.icon_module = 'LiquidDiffract.resources.icons'
        with importlib_resources.path(self.icon_module, 'gs_icon.png') as path:
            self.setWindowIcon(QIcon(str(path)))
        self.setGeometry(0, 0, self.width, self.height)

        self.menu_bar = self.menuBar()
        self.tools_menu = self.menu_bar.addMenu('&Tools')
        self.help_menu = self.menu_bar.addMenu('&Help')

        with importlib_resources.path(self.icon_module, 'config.png') as path:
            self.preferences_action = QAction(QIcon(str(path)),
                                              'Additional Preferences...', self)

        with importlib_resources.path(self.icon_module, 'browser.png') as path:
            self.documentation_action = QAction(QIcon(str(path)),
                                                self.title + ' Documentation', self)

        with importlib_resources.path(self.icon_module, 'info.png') as path:
            self.about_action = QAction(QIcon(str(path)), 'About', self)

        self.tools_menu.addAction(self.preferences_action)
        self.help_menu.addAction(self.documentation_action)
        self.help_menu.addAction(self.about_action)
        self.preferences_action.triggered.connect(self.call_preferences_dialog)
        self.about_action.triggered.connect(self.call_about_dialog)
        self.documentation_action.triggered.connect(self.open_docs)

        self.table_widget = MainContainer(self)
        self.setCentralWidget(self.table_widget)
        self.set_default_preferences()
        self.showMaximized()

    def set_default_preferences(self):
        self.preferences = {'append_log_mode': 1,
                            'data_units': 0,
                            'window_length': 5,
                            'poly_order': 3,
                            'rescale_AL': 1,
                            'fft_N': 12,
                            'mod_func_mode': 0,
                            'op_method': 'L-BFGS-B',
                            'minimisation_options':
                                {'disp': 0,
                                 'maxiter': 15000,
                                 'maxfun': 15000,
                                 'ftol': 2.22e-8,
                                 'gtol': 1e-10},
                            'global_minimisation': 0,
                            'global_min_options':
                                {'disp': 0,
                                 'niter': 100,
                                 'T': 1.0,
                                 'stepsize': 0.01,
                                 'interval': 50},
                            'xray_weight_mode': 0
                            }
        self.set_preferences()

    def call_preferences_dialog(self):
        self.preferences_dialog = utility.PreferencesDialog(self.preferences)
        self.preferences_dialog.fft_check_signal.connect(self.check_fft_N)
        # set window icon
        with importlib_resources.path(self.icon_module, 'gs_icon.png') as path:
            self.preferences_dialog.setWindowIcon(QIcon(str(path)))

        if self.preferences_dialog.exec_() == utility.PreferencesDialog.Accepted:
            self.preferences = self.preferences_dialog.get_preferences()
            # Set preferences in OptimUI
            self.set_preferences()
            # Call method plot_data to update data treatment
            self.table_widget.optim_ui.plot_data()

    def set_preferences(self):
        # Set data and minimisation preferences
        self.table_widget.optim_ui.mod_func_mode = self.preferences['mod_func_mode']
        self.table_widget.optim_ui.op_method = self.preferences['op_method']
        self.table_widget.optim_ui.minimisation_options = self.preferences['minimisation_options']
        self.table_widget.optim_ui.global_minimisation = self.preferences['global_minimisation']
        self.table_widget.optim_ui.global_min_options = self.preferences['global_min_options']
        self.table_widget.optim_ui.append_log_mode = self.preferences['append_log_mode']
        self.table_widget.structure_ui.append_log_mode = self.preferences['append_log_mode']
        self.table_widget.optim_ui.window_length = self.preferences['window_length']
        self.table_widget.optim_ui.poly_order = self.preferences['poly_order']
        # Set N for FFT
        self.table_widget.bkg_ui.fft_N = self.preferences['fft_N']
        self.table_widget.optim_ui.fft_N = self.preferences['fft_N']
        self.table_widget.results_ui.fft_N = self.preferences['fft_N']
        # Set data units for file loading
        self.table_widget.bkg_ui.data_units = self.preferences['data_units']
        # Set preferences for rescaling Ashcroft-Langreth S(Q)/g(r) plots
        self.table_widget.results_ui.rescale_AL = self.preferences['rescale_AL']
        # Set preferences for x-ray weight mode (0: calc Kp at Q=0, 1: calc Q-dependent Kp)
        self.table_widget.structure_ui.xray_weight_mode = self.preferences['xray_weight_mode']

    def call_about_dialog(self):
        self.about_dialog = utility.AboutDialog()
        self.about_dialog.exec_()

    def open_docs(self):
        webbrowser.open_new('https://github.com/bjheinen/LiquidDiffract')

    def check_fft_N(self):
        try:
            _qmax = self.table_widget.bkg_ui.data['data_raw_x'][-1]
        except IndexError:
            try:
                _qmax = self.table_widget.bkg_ui.data['bkg_raw_x'][-1]
            except IndexError:
                self.preferences_dialog.fft_check_result = 0
                return
        _fft_check = self.preferences_dialog.fft_check
        _dx = (self.table_widget.bkg_ui.data['data_raw_x'][1]
               - self.table_widget.bkg_ui.data['data_raw_x'][0])
        if _qmax > ((2**_fft_check / 2) * _dx):
            self.preferences_dialog.fft_check_result = 1


class MainContainer(QWidget):

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.setStyleSheet('QTabWidget::pane { border: 0; } \
                            QTabBar {font-size: 11pt; text-align: center} \
                            QTabWidget::pane {border-top: 4px solid #444; }')
        self.layout = QVBoxLayout(self)
        self.tabs = QTabWidget()
        self.bkg_ui = bkg_ui.BkgUI(self)
        self.optim_ui = optim_ui.OptimUI(self)
        self.results_ui = results_ui.ResultsUI(self)
        self.structure_ui = structure_ui.StructureUI(self)
        self.tabs.addTab(self.bkg_ui, 'Background Subtraction')
        self.tabs.addTab(self.optim_ui, 'Refine Structure Factor')
        self.tabs.addTab(self.results_ui, 'Calculate PDF')
        self.tabs.addTab(self.structure_ui, 'Structural Information')
        self.layout.addWidget(self.tabs)
        self.setLayout(self.layout)

        self.create_signal_tab_links()

    def create_signal_tab_links(self):
        self.bkg_ui.plots_changed.connect(self.bkg_plots_changed_slot)
        self.optim_ui.results_changed.connect(self.results_changed_slot)
        self.optim_ui.results_cleared.connect(self.results_cleared_slot)
        self.bkg_ui.file_name_changed.connect(self.update_filename)
        self.optim_ui.optim_config_widget.optim_results_gb.bkg_scale_copy_btn.clicked.connect(self.copy_bkg_scale_result)

    def bkg_plots_changed_slot(self):
        # Clear data from Optim UI (as S(Q) etc. need to be recalculated)
        self.optim_ui.data = {'uncorrected_y': np.asarray([]), 'bkg_y': np.asarray([]), 'bkg_scale': None,
                              'cor_x': np.asarray([]), 'cor_y': np.asarray([]),
                              'cor_x_cut': np.asarray([]), 'cor_y_cut': np.asarray([]),
                              'sq_x': np.asarray([]), 'sq_y': np.asarray([]),
                              'dr_x': np.asarray([]), 'dr_y': np.asarray([]),
                              'rescaled_cor_y_cut': np.asarray([]),
                              'mod_func': 'None', 'qmin': None, 'qmax': None
                              }
        # Pass data to Optim UI
        self.optim_ui.data['cor_x'] = self.bkg_ui.data['cor_x']
        self.optim_ui.data['cor_y'] = self.bkg_ui.data['cor_y']
        if self.bkg_ui.bkg_config_widget.bkg_subtract_gb.isChecked() and self.bkg_ui.data['bkg_y'].size:
            self.optim_ui.data['uncorrected_y'] = self.bkg_ui.data['data_y']
            self.optim_ui.data['bkg_y'] = self.bkg_ui.data['bkg_y']
            self.optim_ui.data['bkg_scale'] = self.bkg_ui.bkg_config_widget.bkg_subtract_gb.scale_sb.value()
        self.optim_ui.plot_data()

    def results_changed_slot(self):
        # Clear old s(q), g(r), rdf(r)
        self.results_ui.clear_data()
        # Set rho
        if (self.optim_ui.optim_config_widget.optim_options_gb.opt_check.isChecked()) & (self.optim_ui.optim_config_widget.optim_options_gb.isChecked()):
            self.results_ui.data['rho'] = self.optim_ui.data['refined_rho']
        else:
            self.results_ui.data['rho'] = np.float(self.optim_ui.optim_config_widget.composition_gb.density_input.text())
        if self.optim_ui.optim_config_widget.optim_options_gb.isChecked():
            self.results_ui.data['int_func'] = self.optim_ui.data['impr_int_func']
            self.results_ui.data['sq_x'] = self.optim_ui.data['impr_iq_x']
        else:
            self.results_ui.data['int_func'] = self.optim_ui.data['int_func']
            self.results_ui.data['sq_x'] = self.optim_ui.data['iq_x']
        self.results_ui.data['composition'] = self.optim_ui.optim_config_widget.composition_gb.get_composition_dict()
        self.results_ui.data['mod_func'] = self.optim_ui.data['mod_func']
        self.results_ui.data['window_start'] = self.optim_ui.data['window_start']
        self.results_ui.data['sq_method'] = self.optim_ui.data['sq_method']
        self.results_ui.plot_data()

        # Clear data from structure ui
        self.structure_ui.clear_data()

        self.structure_ui.data['rdf_x'] = self.results_ui.data['rdf_x']
        self.structure_ui.data['rdf_y'] = self.results_ui.data['rdf_y']
        self.structure_ui.data['tr_x'] = self.results_ui.data['tr_x']
        self.structure_ui.data['tr_y'] = self.results_ui.data['tr_y']
        self.structure_ui.data['sq_x'] = self.results_ui.data['sq_x']
        self.structure_ui.data['rho'] = self.results_ui.data['rho']
        self.structure_ui.data['composition'] = self.results_ui.data['composition']

        self.structure_ui.update_obj_fun()
        self.structure_ui.set_atoms()
        self.structure_ui.set_weights()
        self.structure_ui.structure_plot_widget.toggle_fit_limits(True,
                                                                  obj_x=self.structure_ui.data['rdf_x'],
                                                                  sq_x=self.structure_ui.data['sq_x'])
        self.structure_ui.plot_data()

    def results_cleared_slot(self):
        self.results_ui.clear_data()
        self.results_ui.results_plot_widget.update_plots(self.results_ui.data)
        
        self.structure_ui.clear_data()
        self.structure_ui.structure_plot_widget.clear_plots(_clear_all=True)

    def update_filename(self):
        _base_name, _ext = os.path.splitext(self.bkg_ui.data_file)
        self.results_ui.base_filename = _base_name
        self.optim_ui.base_filename = _base_name
        self.structure_ui.base_filename = _base_name
        self.optim_ui.filename_ext = _ext
        self.structure_ui.filename_ext = _ext

    def copy_bkg_scale_result(self):
        _bkg_scale_result = self.optim_ui.optim_config_widget.optim_results_gb.bkg_scale_output.text()
        self.bkg_ui.bkg_config_widget.bkg_subtract_gb.scale_sb.setValue(np.float64(_bkg_scale_result))


from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the version number from version.py
version = {}
with open(path.join(here, 'LiquidDiffract/version.py')) as fp:
    exec(fp.read(), version)


# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='LiquidDiffract',
    version=version['__version__'],
    description='Liquid structure/total X-ray scattering analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='GPL',
    author='Benedict J Heinen',
    author_email='benedict.heinen@gmail.com',
    url='https://github.com/bjheinen/LiquidDiffract',
    packages=find_packages(),
    python_requires='>=3.5',
    install_requires=['numpy', 'scipy', 'PyQt5', 'pyqtgraph', 'lmfit',
                      "importlib_resources ; python_version<'3.7'"],
    package_data={'LiquidDiffract': ['resources/*', 'resources/icons/*',
                                     'resources/hubbel_compton/*',
                                     'resources/docs/*',
                                     'scripts/*']},
    zip_safe=True,
    entry_points={'console_scripts':
                  ['LiquidDiffract=LiquidDiffract.LiquidDiffract:main']}
)
